// Copyright 1996, Marimba Inc. All Rights Reserved.
// @(#)TableWidget.java, 1.49, 01/18/97

package marimba.gui;

import java.awt.*;
import java.util.*;

import marimba.persist.*;
import marimba.util.*;
import marimba.io.*;

/**
 * A table widget for multiple columns and rows of data.
 *
 * Every cell can contain the following objects: Widget, Boolean,
 * Character, Double, Float, Integer, Long and String.
 * A Widget is placed on the correct position and its width
 * is adjusted to the width of the column. All other objects
 * are just painted at the correct location.
 *
 * This TableWidget supports sorting on columns, dragging columns
 * around, row selections, resizing of columns and both horizontal
 * and vertical scrollbars of course.
 *
 * @author	Klaas Waslander
 * @version 	1.49, 01/18/97
 */ 
public class TableWidget extends ScrollingContainerWidget {
    final static Font defaultHdrFont = new Font("Dialog", Font.PLAIN, 10);

    /**
     * The possible options for the column header style,
     * which are none, plain, outline and raised.
     * @see #getColHeadersOptions
     * @see #colHeaders
     */
    public static Options colHeadersOptions = new Options();
    static {
	colHeadersOptions.add("none", NONE);
	colHeadersOptions.add("plain", PLAIN);
	colHeadersOptions.add("outline", OUTLINE);
	colHeadersOptions.add("raised", RAISED);
    }

    /**
     * The possible options for sorting,
     * which are none, ascending, descending, both.
     * @see #getSortModeOptions
     * @see #sortMode
     */
    public static Options sortModeOptions = new Options();
    static {
	sortModeOptions.add("none", NONE);
	sortModeOptions.add("ascending", ASCENDING);
	sortModeOptions.add("descending", DESCENDING);
	sortModeOptions.add("both", BOTH);
    }

    /**
     * The possible options for selecting,
     * which are none, exclusive or multiple.
     * @see #getSelectModeOptions
     * @see #selectMode
     */
    public static Options selectModeOptions = new Options();
    static {
	selectModeOptions.add("none", NONE);
	selectModeOptions.add("exclusive", EXCLUSIVE);
	selectModeOptions.add("multiple", MULTIPLE);
    }

    /**
     * The current option for sorting.
     * @see #getSortMode
     * @see #setSortMode
     * @see #sortModeOptions
     */
    public int sortMode = BOTH;

    /**
     * Determines whether the user can 'unsort' the data
     * by clicking the column header.
     * @see #getUnsort
     * @see #setUnsort
     */
    public boolean unsort = true;

    /**
     * The current sorting can be drawn in the headers using
     * an appropriate symbol.
     * @see #getSortSymbol
     * @see #setSortSymbol
     */
    public boolean sortSymbol = true;

    /**
     * The font that has to be used for the header.
     * @see #getHdrFont
     * @see #setHdrFont
     */
    public Font  hdrFont = defaultHdrFont;

    /**
     * The type of selections that can be made:
     * none, exclusive or multiple selections.
     * @see #getSelectMode
     * @see #setSelectMode
     * @see #selectModeOptions
     */
    public int selectMode = EXCLUSIVE;

    /**
     * The columns can be moved by dragging them around.
     * This is never possible if colHeaders is set to none.
     * @see #getColumnDrag
     * @see #setColumnDrag
     */
    public boolean  columnDrag = true;

    /**
     * The columns can be separated by a line.
     * @see #getColumnLine
     * @see #setColumnLine
     */
    public boolean  columnLine = false;

    /**
     * The columns can be resizable in the header area.
     * This is never possible if colHeaders is set to none.
     * @see #getColumnResize
     * @see #setColumnResize
     */
    public boolean  columnResize = true;

    /**
     * The current option for column headers.
     * @see #getColHeaders
     * @see #setColHeaders
     * @see #colHeaderOptions
     */
    public int  colHeaders = RAISED;

    /**
     * The headers can be filled or transparent.
     * @see #getFillHeaders
     * @see #setFillHeaders
     */
    public boolean  fillHeaders = false;

    /**
     * The widgets can be sized to the cell size.
     * @see #getSizeToCell
     * @see #setSizeToCell
     */
    public boolean sizeToCell = false;

    /**
     * The number of pixels space at the top and bottom
     * of each cell.
     * @see #getRowPadding
     * @see #setRowPadding
     */
    public int rowPadding = 4;

    /**
     * The number of pixels space at the left and right
     * of each cell.
     * @see #getColPadding
     * @see #setColPadding
     */
    public int colPadding = 4;

    /**
     * The display index of the current row.
     * @see #getCurrentRow
     * @see #setCurrentRow
     * @see #getCurrentRowIndex
     * @see #setCurrentRowIndex
     */
    public int  currentRowIndex = -1;

    /**
     * The information about the columns is stored in this array.
     * It contains a TableColumn for each column.
     * @see #getColumn
     */
    protected TableColumn  columns[] = new TableColumn[4];

    /**
     * The information about the rows is stored in this array.
     * It contains a TableRow for each row with the data of that row.
     * @see #getRow
     */
    protected TableRow  rows[] = new TableRow[4];

    /**
     * The currently selected rows, this vector contains the
     * actual index of every selected row.
     * @see #getSelectedRows
     * @see #getSelectedRowIndexes
     * @see #rowIsSelected
     * @see #rowIsSelectedAt
     */
    protected int[]  selectedRows = new int[0];

    /**
     * Array for mapping column display index to a column index.
     * The array index is the display index, the item is the actual column
     * index. Mapping the other way around can be done by asking a column
     * for its 'displayIndex'.
     * @see #getColumnIndex
     */
    protected int  columnMap[] = new int[4];

    /**
     * Array for mapping row display index to an actual row index.
     * The array index is the display index, the item is the actual row
     * index. Mapping the other way around can be done by asking a row
     * for its 'displayIndex'.
     * @see #getRowIndex
     */
    protected int  rowMap[] = new int[4];

    /**
     * The number of rows.
     * @see #rowCount
     */
    protected int  rowCount = 0;

    /**
     * The number of columns.
     * @see #columnCount
     */
    protected int  columnCount = 0;

    /**
     * The current sort method, ascending/descending/none.
     * @see #getSortMethod
     */
    protected int  sortMethod = NONE;

    /**
     * The current column index on which is being sorted, which
     * is the actual index and not the display index.
     * @see #getSortIndex
     */
    protected int  sortIndex = -1;


    /** the column button currently being pressed */
    int  downButton = 0;

    /** the column before which the current column is being moved */
    int  moveColumn = -1;

    /** indicates whether a button is down */
    boolean  down = false;

    /** the last x position of the mouse, for handling events */
    int  oldX;


    /**
     * Constructor.
     */
    public TableWidget() {
	super();
	corner = true;
	horzmode = OPTIONAL;
	vertmode = OPTIONAL;
    }

    /**
     * Get the properties of this widget.
     */
    public void getProperties(PropertyList list) {
	super.getProperties(list);
	//list.setBoolean("exclusive", exclusive, true);
	list.setFont("hdrfont", hdrFont, defaultHdrFont);
	list.setBoolean("fillHeaders", fillHeaders, false);
	list.setOption("colHeaders", colHeadersOptions, colHeaders, RAISED);
	list.setOption("selectMode", selectModeOptions, selectMode, EXCLUSIVE);

	// persistify rows and columns
	list.setObjectArray("rowArray", rowCount, rows);
	list.setObjectArray("columnArray", columnCount, columns);
	
	// persistify column mapping
	FastOutputStream  colbuf = new FastOutputStream();
	for (int i = 0; i < columnCount; i++) {
	    colbuf.writeInt(columnMap[i]);
	}
	list.setByteArray("columnMap", colbuf.getByteArray(), null);

	// persistify row mapping
	FastOutputStream  rowbuf = new FastOutputStream();
	for (int i = 0; i < rowCount; i++) {
	    rowbuf.writeInt(rowMap[i]);
	}
	list.setByteArray("rowMap", rowbuf.getByteArray(), null);

	// the rest
	list.setString("columns", getColumnHeaders(), "");
	list.setBoolean("columnDrag", columnDrag, true);
	list.setBoolean("columnLine", columnLine, false);
	list.setBoolean("columnResize", columnResize, true);
	list.setInteger("rowPadding", rowPadding, 4);
	list.setInteger("colPadding", colPadding, 4);
	list.setBoolean("sizeToCell", sizeToCell, false);
	list.setOption("sortMode", sortModeOptions, sortMode, BOTH);
	list.setBoolean("unsort", unsort, true);
	list.setBoolean("sortSymbol", sortSymbol, true);
    }

    /**
     * Set the properties of this widget.
     */
    public void setProperties(PropertyList list) {
	super.setProperties(list);
	//exclusive = list.getBoolean("exclusive", true);
	hdrFont = list.getFont("hdrfont", defaultHdrFont);
	fillHeaders = list.getBoolean("fillHeaders", false);
	colHeaders = list.getOption("colHeaders", colHeadersOptions, RAISED);
	selectMode = list.getOption("selectMode", selectModeOptions, EXCLUSIVE);

	// REMIND: backward compatibility introduced at Jan. 9 1997
	boolean  exclusive = list.getBoolean("exclusive", true);
	if (!exclusive) {
	    selectMode = MULTIPLE;
	}

	// get rows
	Object  temp[] = list.getObjectArray("rowArray", new TableRow[4]);
	rows = new TableRow[temp.length];
	rowCount = temp.length;
	System.arraycopy(temp, 0, rows, 0, temp.length);

	// get columns
	temp = list.getObjectArray("columnArray", new TableColumn[4]);
	columns = new TableColumn[temp.length];
	columnCount = temp.length;
	System.arraycopy(temp, 0, columns, 0, temp.length);

	// get column mappings
	columnMap = new int[columnCount];
	byte[]  colMap = list.getByteArray("columnMap", null);
	FastInputStream  colbuf = new FastInputStream(colMap);
	for (int i = 0; i < columnCount; i++) {
	    columnMap[i] = colbuf.readInt();
	}

	// get row mappings
	rowMap = new int[rowCount];
	byte[]  rMap = list.getByteArray("rowMap", null);
	FastInputStream  rowbuf = new FastInputStream(rMap);
	for (int i = 0; i < rowCount; i++) {
	    rowMap[i] = rowbuf.readInt();
	}

	// the rest
	setColumnHeaders(list.getString("columns",""));
	columnDrag = list.getBoolean("columnDrag", true);
	columnLine = list.getBoolean("columnLine", false);
	columnResize = list.getBoolean("columnResize", true);
	rowPadding = list.getInteger("rowPadding", 4);
	colPadding = list.getInteger("colPadding", 4);
	sizeToCell = list.getBoolean("sizeToCell", false);
	setSortMode(list.getOption("sortMode", sortModeOptions, BOTH));
	unsort = list.getBoolean("unsort", true);
	sortSymbol = list.getBoolean("sortSymbol", true);

	invalidate();
	repaint();
	content.repaint();
    }

    /**
     * Get the column headers as a string. This is used in the
     * getProperties method and if you pass this string to
     * setColumnHeaders it will result in the correct headers.
     * @see #setColumnHeaders
     */
    public String getColumnHeaders() {
	FontMetrics  hdrMetrics = getFontMetrics(hdrFont);
	StringBuffer buf = new StringBuffer();
	int  colCount = columnCount();
	for (int i = 0; i < colCount; i++) {
	    if (i > 0) {
		buf.append(',');
	    }
	    TableColumn  col = getColumn(i);
	    int  colWidth = hdrMetrics.stringWidth(col.label)+10;
	    
	    // replace '/' with '//' in column label
	    String  label = new String(col.label);
	    int  lastSlash = 0;
	    int  slashIndex = label.indexOf("/", lastSlash);
	    while (slashIndex >= 0) {
		StringBuffer  slashBuf = new StringBuffer();
		slashBuf.append(label.substring(0, slashIndex));
		slashBuf.append("//");
		slashBuf.append(label.substring(slashIndex+1, label.length()));
		label = slashBuf.toString();

		lastSlash = slashIndex + 2;
		slashIndex = label.indexOf("/", lastSlash);
	    }

	    // append label to string and add size if necessary	    
	    buf.append(label);
	    if (col.size != colWidth) {
		buf.append('/');
		buf.append(col.size);
	    }
	}
	return buf.toString();
    }


    /**
     * Base the columns on these column headers. Names are changed if
     * necessary and columns removed/added if appropriate.
     * Column names must be separated by a comma, each column name can be
     * given a width using a slash. For example: "one/20,two,three/50".
     * If you want a column name with a slash in it, you have to supply
     * two slashes. For example: "12//3/20, 5//2".
     * @see #getColumnHeaders
     */
    public synchronized void setColumnHeaders(String str) {
	// do nothing if one of the parameters is incorrect
	if (str == null) {
	    paramError("setColumnHeaders(String)", "null string specified");
	    return;
	}

	// base columns on given string
	if (str.length() > 0) {
	    FontMetrics  hdrMetrics = getFontMetrics(hdrFont);
	    StringTokenizer  tokenizer = new StringTokenizer(str, ",");

	    int  colIndex;
	    for (colIndex = 0; tokenizer.hasMoreTokens(); colIndex++) {
		// get the existing column and the new label and its width
		TableColumn  currentCol = (colIndex < columnCount) ? getColumn(colIndex) : null;
		String  colLabel = tokenizer.nextToken();

		// deal with two slashes for getting a slash in the column header
		int  lastSkip = 0;
		int  skipSlash = colLabel.indexOf("//", lastSkip);
		while (skipSlash >= 0) {
		    StringBuffer  buf = new StringBuffer();
		    buf.append(colLabel.substring(0, skipSlash));
		    buf.append('/');
		    buf.append(colLabel.substring(skipSlash+2, colLabel.length()));
		    colLabel = buf.toString();

		    lastSkip = skipSlash + 1;
		    skipSlash = colLabel.indexOf("//", lastSkip);
		}

		// determine width of the column
		int  colWidth = hdrMetrics.stringWidth(colLabel)+10;

		int  widthIndex = colLabel.indexOf('/', lastSkip);
		if (widthIndex >= 0) {
		    // only adjust column width if there is something after slash
		    String  widthString = colLabel.substring(widthIndex + 1);
		    if (widthString.length() > 0) {
			try {
			    colWidth = Integer.parseInt(widthString);
			} catch (NumberFormatException e) {
			    colWidth = hdrMetrics.stringWidth(colLabel.substring(0, widthIndex))+10;
			}
		    } else {
			colWidth = hdrMetrics.stringWidth(colLabel.substring(0, widthIndex))+10;
		    }
		    colLabel = colLabel.substring(0, widthIndex);
		}

		// compare this label and width with the current column
		if (currentCol != null) {
		    currentCol.label = colLabel;
		    changeColumnWidth(colIndex, Math.max(1, colWidth) - currentCol.size);
		} else {
		    addColumn(colLabel, colWidth);
		}
	    }

	    // remove remaining columns
	    for (int remove = columnCount-1; remove >= colIndex; remove--) {
		removeColumnAt(remove);
	    }

	    invalidate();
	    repaint();
	    content.repaint();
	} else {
	    removeAllColumns();
	}
    }

    /**
     * Allocate content container.
     */
    protected Widget newContent() {
	TableContent  result = new TableContent();
	result.lineMode = LOWERED;
	return result;
    }

    /**
     * Create a new scrollbar with all the necessary properties.
     */
    public ScrollbarWidget newScrollbar(int orientation) {
	ScrollbarWidget  result = super.newScrollbar(orientation);
	result.setFilled(fillMode==SCROLLBAR || fillMode==FILLED);
	return result;
    }

    /**
     * Get the index of the first selected row.
     */
    public int getIntegerValue() {
	TableRow  first = (TableRow)getSelectedRows().firstElement();
	return getRowIndex(first.displayIndex);
    }

    /**
     * Get the index of the first selected row.
     */
    public Object getValue() {
	return new Integer(getIntegerValue());
    }

    /**
     * Select a row by specifying a Number, String or
     * TableRow object.
     * @see #selectedRows
     */
    public void setValue(Object value) {
	if (value instanceof Number) {
	    setValue(((Number)value).intValue());
	} else if (value instanceof String) {
	    setValue(Integer.parseInt((String)value));
	} else if (value instanceof TableRow) {
	    selectRow((TableRow)value);
	} else if (value == null) {
	    deselectAllRows();
	}
    }

    /**
     * Select a row.
     * @see #selectedRows
     */
    public void setValue(int rowIndex) {
	selectRowAt(rowIndex);
    }

    /**
     * Get the possible options for the column headers.
     * @see #colHeadersOptions
     */
    public Options getColHeadersOptions() {
	return colHeadersOptions;
    }

    /**
     * Get the possible options for the sort mode.
     * @see #sortModeOptions
     */
    public Options getSortModeOptions() {
	return sortModeOptions;
    }

    /**
     * Get the current sort mode.
     * @see #sortMode
     */
    public int getSortMode() {
	return sortMode;
    }

    /**
     * Set the sort mode to be none, ascending, descending or both.
     * @see #sortMode
     */
    public void setSortMode(int sortMode) {
	if (this.sortMode != sortMode) {
	    this.sortMode = sortMode;
	    sortMethod = NONE;
	}
    }

    /**
     * Check whether the user can unsort the data when
     * clicking on a header.
     * @see #unsort
     */
    public boolean getUnsort() {
	return unsort;
    }

    /**
     * (Dis)allow unsorting the data using the headers.
     * @see #unsort
     */
    public void setUnsort(boolean unsort) {
	this.unsort = unsort;
    }

    /**
     * Check whether a sort symbol is shown in a header.
     * @see #sortSymbol
     */
    public boolean getSortSymbol() {
	return sortSymbol;
    }

    /**
     * Show or hide the sort symbol for showing the current
     * sorting column column and its way of sorting.
     * @see #sortSymbol
     */
    public void setSortSymbol(boolean sortSymbol) {
	if (this.sortSymbol != sortSymbol) {
	    this.sortSymbol = sortSymbol;
	    repaint();
	}
    }

    /**
     * Get the possible options for the select mode.
     * @see #selectModeOptions
     */
    public Options getSelectModeOptions() {
	return selectModeOptions;
    }

    /**
     * Check the select mode.
     * @see #selectMode
     */
    public int getSelectMode() {
	return selectMode;
    }

    /**
     * Set the select mode to none, exclusive or multiple.
     * @see #selectMode
     */
    public void setSelectMode(int selectMode) {
	this.selectMode = selectMode;
    }

    /**
     * Get the font for the header.
     * @see #hdrFont
     */
    public Font getHdrFont() {
	return hdrFont;
    }

    /**
     * Set the font for the header.
     * @see #hdrFont
     */
    public void setHdrFont(Font hdrFont) {
	if (hdrFont != null && !hdrFont.equals(this.hdrFont)) {
	    this.hdrFont = hdrFont;
	    invalidate();
	    repaint();
	}
    }

    /**
     * Check whether the columns can be moved by dragging them around.
     * @see #columnDrag
     */
    public boolean getColumnDrag() {
	return columnDrag;
    }

    /**
     * Set dragging of columns to true or false.
     * @see #columnDrag
     */
    public void setColumnDrag(boolean columnDrag) {
	this.columnDrag = columnDrag;
    }

    /**
     * Check whether the columns are separated by a line.
     * @see #columnLine
     */
    public boolean getColumnLine() {
	return columnLine;
    }

    /**
     * Show/hide the lines between columns.
     * @see #columnLine
     */
    public void setColumnLine(boolean columnLine) {
	if (this.columnLine != columnLine) {
	    this.columnLine = columnLine;
	    content.repaint();
	}
    }

    /**
     * Check whether the columns are resizable from the header area.
     * @see #columnResize
     */
    public boolean getColumnResize() {
	return columnResize;
    }

    /**
     * Disable or enable column resizing in the header area.
     * @see #columnResize
     */
    public void setColumnResize(boolean columnResize) {
	this.columnResize = columnResize;
    }

    /**
     * Get the current option for the colHeaders.
     * @see #colHeaders
     */
    public int getColHeaders() {
	return colHeaders;
    }

    /**
     * Set the colHeaders to be none, plain, outline or raised.
     * @see #colHeaders
     */
    public void setColHeaders(int colHeaders) {
	if (colHeaders != this.colHeaders) {
	    this.colHeaders = colHeaders;
	    layout();
	}
    }

    /**
     * Check whether the headers are filled.
     * @see #fillHeaders
     */
    public boolean getFillHeaders() {
	return fillHeaders;
    }

    /**
     * Fill the headers area or let the area be transparent.
     * @see #fillHeaders
     */
    public void setFillHeaders(boolean fillHeaders) {
	if (this.fillHeaders != fillHeaders) {
	    this.fillHeaders = fillHeaders;
	    repaint();
	}
    }

    /**
     * Check whether the widgets are sized to the cell size.
     * @see #sizeToCell
     */
    public boolean getSizeToCell() {
	return sizeToCell;
    }

    /**
     * Size the widgets to the cell size automatically or not.
     * @see #sizeToCell
     */
    public void setSizeToCell(boolean sizeToCell) {
	if (this.sizeToCell != sizeToCell) {
	    this.sizeToCell = sizeToCell;
	    layout();
	}
    }

    /**
     * Get the number of columns.
     * @see #columnCount
     */
    public int columnCount() {
	return columnCount;
    }

    /**
     * Get the number of rows.
     * @see #rowCount
     */
    public int rowCount() {
	return rowCount;
    }

    /**
     * Get the number of pixels at the top and
     * bottom of each cell.
     * @see #rowPadding
     */
    public int getRowPadding() {
	return rowPadding;
    }

    /**
     * Set the number of pixels at the top and
     * bottom of each cell.
     * @see #rowPadding
     */
    public void setRowPadding(int rowPadding) {
	if (rowPadding != this.rowPadding) {
	    this.rowPadding = rowPadding;
	    layout();
	    content.repaint();
	}
    }

    /**
     * Get the number of pixels at the left and
     * right of each cell.
     * @see #colPadding
     */
    public int getColPadding() {
	return colPadding;
    }

    /**
     * Set the number of pixels at the left and
     * right of each cell.
     * @see #colPadding
     */
    public void setColPadding(int colPadding) {
	if (colPadding != this.colPadding) {
	    this.colPadding = colPadding;
	    layout();
	    content.repaint();
	}
    }

    /**
     * Get the width of all the columns together.
     */
    public int columnsWidth() {
	int  result = 0;    
	for (int index = 0; index < columnCount; index++) {
	    result += getColumn(index).size;
	}
	return result;
    }

    /**
     * Get the height of all the rows together.
     */
    public int rowsHeight() {
	return ((TableContent)content).getRowsHeight();
    }

    /**
     * Set the column width of the column with the given actual index
     * and adjust the x positions of the other columns.
     */
    public void setColumnWidth(int index, int width) {
	// do nothing if index or width is incorrect
	if (index >= columnCount || index < 0) {
	    paramError("setColumnWidth(int,int)", "no column at position " + index);
	    return;
	}
	width = Math.max(2, width);
	TableColumn  col = getColumn(index);
	int  widthChange = col.size - width;
	if (widthChange != 0) {
	    col.size = width;
	    for (int colNumber = col.displayIndex+1; colNumber < columnCount; colNumber++) {
		col = getColumn(getColumnIndex(colNumber));
		col.pos -= widthChange;
	    }
	    invalidate();
	    repaint();
	}
    }

    /**
     * Change the column width of the column with the given actual index.
     */
    public void changeColumnWidth(int index, int widthChange) {
	// do nothing if index or change of width is incorrect
	if (index >= columnCount || index < 0) {
	    paramError("changeColumnWidth(int,int)", index + " and " + widthChange);
	    return;
	}
	setColumnWidth(index, getColumn(index).size + widthChange);
    }

    /**
     * Get the TableColumn with info about the column
     * with the given index, not dealing with columns
     * being switched.
     */
    public TableColumn getColumn(int columnIndex) {
	// do nothing if one of the parameters is incorrect
	if (columnIndex < 0 || columnIndex >= columnCount) {
	    paramError("getColumn(int)", "column number " + columnIndex + " does not exist");
	    return null;
	}
	return columns[columnIndex];
    }

    /**
     * Get the TableRow with info about the row
     * with the given index, not dealing with rows
     * being sorted.
     */
    public TableRow getRow(int index) {
	// do nothing if one of the parameters is incorrect
	if (index < 0 || index >= rowCount) {
	    paramError("getRow(int)", "row number " + index + " does not exist");
	    return null;
	}
	return rows[index];
    }

    /**
     * Get the data of the column with the given index.
     * It is returned in the original unsorted order.
     */
    public Vector getColumnData(int columnIndex) {
	// do nothing if one of the parameters is incorrect
	if (columnIndex < 0 || columnIndex >= columnCount) {
	    paramError("getColumnData(int)", "column number " + columnIndex + " does not exist");
	    return null;
	}
	Vector  result = new Vector();
	for (int rowIndex = 0; rowIndex < rowCount; rowIndex++) {
	    result.addElement(getCell(columnIndex, rowIndex));
	}
	return result;
    }

    /**
     * Get the data of the row with the given index.
     * It is returned in the original order, not the current display order.
     */
    public Vector getRowData(int index) {
	// do nothing if one of the parameters is incorrect
	if (index < 0 || index >= rowCount) {
	    paramError("getRowData(int)", "row number " + index + " does not exist");
	    return null;
	}
	return getRow(index).data;
    }

    /**
     * Get the data in the cell that corresponds with the
     * given (column, row) pair. It uses the original column and row
     * index, so it does not take account of row sorting and a
     * different display order of the columns.
     * @see #setCell
     */
    public Object getCell(int colIndex, int rowIndex) {
	// do nothing if one of the parameters is incorrect
	if (colIndex < 0 || colIndex >= columnCount || rowIndex < 0 || rowIndex >= rowCount) {
	    paramError("getCell(int,int)", "(colIndex, rowIndex) = ("+colIndex+","+rowIndex+") does not exist");
	    return null;
	}
	Vector  rowData = getRowData(rowIndex);
	return rowData.elementAt(colIndex);
    }

    /**
     * Set the content of the cell that corresponds with
     * the given (column, row) pair. The old content is removed.
     * You can supply null to make the cell empty.
     * @see #getCell
     */
    public void setCell(Object item, int colIndex, int rowIndex) {
	// do nothing if one of the parameters is incorrect
	if (colIndex < 0 || colIndex >= columnCount || rowIndex < 0 || rowIndex >= rowCount) {
	    paramError("setCell(int,int)", "(colIndex, rowIndex) = ("+colIndex+","+rowIndex+") does not exist");
	    return;
	}
	Vector  rowData = getRowData(rowIndex);
	for (int i = rowData.size(); i <= colIndex; i++) {
	    rowData.addElement(null);
	}
	rowData.setElementAt(item, colIndex);

	content.invalidate();
	content.repaint();
    }

    /**
     * Get the label of the given column.
     */
    public String getColumnLabelAt(int index) {
	// do nothing if one of the parameters is incorrect
	if (index > columnCount || index < 0) {
	    paramError("getColumnLabelAt(int)", "column index out of range");
	    return null;
	}
	return getColumn(index).label;
    }

    /**
     * Set the label of the given column.
     */
    public void setColumnLabelAt(String newLabel, int index) {
	// do nothing if one of the parameters is incorrect
	if (index > columnCount || index < 0) {
	    paramError("setColumnLabelAt(String,int)", "column index out of range");
	    return;
	}
	if (newLabel == null) {
	    newLabel = "";
	}
	if (!newLabel.equals(getColumn(index).label)) {
	    getColumn(index).label = newLabel;
	    repaint();
	}
    }

    /**
     * Get the alignment of the given column.
     */
    public int getColumnAlignAt(int index) {
	// do nothing if one of the parameters is incorrect
	if (index > columnCount || index < 0) {
	    paramError("getColumnAlignAt(int)", "column index out of range");
	    return -1;
	}
	return getColumn(index).align;
    }

    /**
     * Set the alignment of the given column to right, left or center.
     */
    public void setColumnAlignAt(int align, int index) {
	// do nothing if one of the parameters is incorrect
	if (align != LEFT && align != RIGHT && align != CENTER) {
	    paramError("setColumnAlignAt(int,int)", "given alignment was not left, right or center");
	    return;
	}
	if (index > columnCount || index < 0) {
	    paramError("setColumnAlignAt(int,int)", "column index out of range");
	    return;
	}
	if (getColumn(index).align != align) {
	    getColumn(index).align = align;
	    repaint();
	}
    }

    /**
     * Append an empty column without header.
     */
    public void addColumn() {
	addColumn(null);
    }

    /**
     * Append an empty column to this table.
     */
    public void addColumn(String header) {
	insertColumnAt(header, columnCount());
    }

    /**
     * Append an empty column with the given width to this table.
     */
    public void addColumn(String header, int width) {
	insertColumnAt(header, columnCount(), width);
    }

    /**
     * Insert an empty column at the given position to this table.
     */
    public void insertColumnAt(String header, int index) {
	insertColumnAt(header, index, getFontMetrics(hdrFont).stringWidth(header)+10);
    }

    /**
     * Insert an empty column at the given position and
     * with the given width to this table.
     * Its actual index and display index will be the same.
     */
    public void insertColumnAt(String header, int index, int width) {
	// do nothing if one of the parameters is incorrect
	if (index > columnCount || index < 0 || width < 0) {
	    paramError("insertColumnAt(String,int,int)");
	    return;
	}
	if (header == null) {
	    header = "";
	}

	// the new x-coordinate of the new column
	int  newX = (index > columnCount()-1) ? columnsWidth() : getColumn(getColumnIndex(index)).pos;

	// grow columns and mappings array if necessary
	if (columnCount + 1 >= columns.length) {
	    TableColumn  newColumns[] = new TableColumn[Math.max(columnCount*2, 4)];
	    System.arraycopy(columns, 0, newColumns, 0, columnCount);
	    columns = newColumns;

	    int  newColumnMap[] = new int[Math.max(columnCount*2, 4)];
	    System.arraycopy(columnMap, 0, newColumnMap, 0, columnCount);
	    columnMap = newColumnMap;
	}

	// add new column at given position and check headers
	System.arraycopy(columns, index, columns, index+1, columnCount-index);
	columns[index] = new TableColumn(header, newX, width, index);
	columnCount++;

	// update column mapping
	System.arraycopy(columnMap, index, columnMap, index+1, columnCount-index);
	columnMap[index] = index;
	for (int i = 0; i < columnCount; i++) {
	    if (i != index && columnMap[i] >= index) {
		columnMap[i]++;
	    }
	}

	// adjust x-coordinates and displayIndex of columns displayed at the right
	for (int i = columnCount-1; i > index; i--) {
	    TableColumn  col = getColumn(getColumnIndex(i));
	    col.pos += width;
	    col.displayIndex++;
	}

	// move the data in the TableRows if it is not the last column
	if (index+1 < columnCount) {
	    for (int i = 0; i < rowCount; i++) {
		Vector  data = getRowData(i);
		// fill with empty elements up to this column if necessary
		for (int fill = data.size()-1; fill < index; fill++) {
		    data.insertElementAt(null, fill);
		}
		data.insertElementAt(null, index);
	    }
	}	

	repaint();
	content.repaint();
	layout();
    }

    /**
     * Remove all columns with their data.
     */
    public void removeAllColumns() {
	columns = new TableColumn[4];
	columnCount = 0;
	columnMap = new int[4];
	removeAllRows();
	repaint();
    }

    /**
     * Remove the column at the given position, which is
     * the actual index and not the display index.
     * The data in that column is also removed.
     */
    public void removeColumnAt(int columnIndex) {
	// do nothing if one of the parameters is incorrect
	if (columnIndex < 0 || columnIndex >= columnCount) {
	    paramError("removeColumnAt(int)", "columnPosition " + columnIndex + " does not exist");
	    return;
	}

	TableColumn  oldCol = getColumn(columnIndex);

	// remove column from columns-array and prefend memory leak
	System.arraycopy(columns, columnIndex+1, columns, columnIndex, columnCount-columnIndex-1);
	columnCount--;
	columns[columnCount] = null;

	// update column mapping
	if (oldCol.displayIndex != columnCount) {
	    System.arraycopy(columnMap, oldCol.displayIndex+1, columnMap, oldCol.displayIndex, columnCount-oldCol.displayIndex);
	}
	for (int i = 0; i < columnCount; i++) {
	    if (columnMap[i] > columnIndex) {
		columnMap[i]--;
	    }
	}

	// adjust x-coordinates and displayIndex of columns displayed at the right
	for (int i = columnCount-1; i >= oldCol.displayIndex; i--) {
	    TableColumn  col = getColumn(getColumnIndex(i));
	    col.pos -= oldCol.size;
	    col.displayIndex--;
	}

	// remove the data from the TableRows
	for (int i = 0; i < rowCount; i++) {
	    Vector  data = getRowData(i);
	    if (data.size() >= columnIndex+1) {
		data.removeElementAt(columnIndex);
	    }
	}

	invalidate();
	repaint();
    }

    /**
     * Get the index of the column that is displayed at the given
     * position. Mapping the display index to the actual index is
     * necessary, since you can only search using the actual index.
     * Mapping from actual index to display index can be done by
     * asking for 'displayIndex' in TableColumn.
     * @see #columnMap
     */
    public int getColumnIndex(int displayIndex) {
	return columnMap[displayIndex];
    }

    /**
     * Move the 'from' column before the 'to' column.
     * The index for both columns is the current display index,
     * so not the actual index.
     */
    public void moveColumn(int fromIndex, int toIndex) {
	// do nothing if one of the parameters is incorrect
	if (fromIndex < 0 || fromIndex >= columnCount) {
	    paramError("moveColumn(int,int)", "columnPosition " + fromIndex + " does not exist");
	    return;
	}
	if (toIndex < 0 || toIndex >= columnCount) {
	    paramError("switchColumns(int,int)", "columnPosition " + toIndex + " does not exist");
	    return;
	}

	TableColumn  fromCol = getColumn(getColumnIndex(fromIndex));
	TableColumn  toCol = getColumn(getColumnIndex(toIndex));
	if (fromIndex < toIndex) {
	    for (int colIndex = fromIndex; colIndex <= toIndex-1; colIndex++) {
		switchColumns(colIndex, colIndex+1);
	    }
	} else if (fromIndex > toIndex) {
	    for (int colIndex = fromIndex; colIndex > toIndex; colIndex--) {
		switchColumns(colIndex, colIndex-1);
	    }
	}
    }

    /**
     * Switch two columns. The index for both columns is the current
     * display index, so not the original index. The display index can
     * be retreived from the TableColumn by asking for its 'displayIndex'.
     */
    public void switchColumns(int colIndex1, int colIndex2) {
	// do nothing if one of the parameters is incorrect
	if (colIndex1 < 0 || colIndex1 >= columnCount) {
	    paramError("switchColumns(int,int)", "columnPosition " + colIndex1 + " does not exist");
	    return;
	}
	if (colIndex2 < 0 || colIndex2 >= columnCount) {
	    paramError("switchColumns(int,int)", "columnPosition " + colIndex2 + " does not exist");
	    return;
	}

	// get the columns, with the first one the left one
	int  smallIndex = Math.min(colIndex1, colIndex2);
	int  largeIndex = Math.max(colIndex1, colIndex2);
	int  actual1 = getColumnIndex(smallIndex);
	int  actual2 = getColumnIndex(largeIndex);
	TableColumn  col1 = getColumn(actual1);
	TableColumn  col2 = getColumn(actual2);

	// switch the display index
	int  col1Display = col1.displayIndex;
	col1.displayIndex = col2.displayIndex;
	col2.displayIndex = col1Display;

	// update the column index mappings
	columnMap[col1.displayIndex] = actual1;
	columnMap[col2.displayIndex] = actual2;

	// switch the x-coordinates
	int  col1Pos = col1.pos;
	col1.pos = col2.pos;
	col2.pos = col1Pos;

	// update the x-coordinates of the columns between
	int  pixelChange = col2.size - col1.size;
	for (int i = smallIndex + 1; i <= largeIndex; i++) {
	    getColumn(getColumnIndex(i)).pos += pixelChange;
	}

	invalidate();
	repaint();
    }

    /**
     * Sort on the given column.
     * @param mode The sortmode, which means ascending or descending.
     */
    public void sortOnColumn(TableColumn sortCol, int mode) {
	// do nothing if no sorting column was given
	if (sortCol == null) {
	    paramError("sortOnColumn(TableColumn)", "null column specified");
	    return;
	}
	sortOnColumnAt(getColumnIndex(sortCol.displayIndex), mode);
    }

    /**
     * Sort on the column at the given actual index,
     * so not a display index.
     * @param mode The sortmode, which means ascending or descending.
     */
    public void sortOnColumnAt(int colIndex, int mode) {
	// do nothing if one of the parameters is incorrect
	if (colIndex < 0 || colIndex >= columnCount) {
	    paramError("sortOnColumnAt(int)", "column position " + colIndex + " does not exist");
	    return;
	}

	// store old current row index
	int  oldActualCurrent = -1;
	if (currentRowIndex >= 0) {
	    oldActualCurrent = rowMap[currentRowIndex];
	}

	// sort rows
	TableRow  sortedRows[] = new TableRow[rowCount];
	System.arraycopy(rows, 0, sortedRows, 0, rowCount);
	QuickSort.sort(sortedRows, 0, rowCount, new Integer(colIndex));

	// update row mapping
	int  newRowMap[] = new int[rowMap.length];
	if (mode == DESCENDING) {
	    for (int i = 0; i < rowCount; i++) {
		TableRow  currentRow = sortedRows[rowCount-1-i];
		newRowMap[i] = getRowIndex(currentRow.displayIndex);
		currentRow.displayIndex = i;
	    }
	} else if (mode == ASCENDING) {
	    for (int i = 0; i < rowCount; i++) {
		TableRow  currentRow = sortedRows[i];
		newRowMap[i] = getRowIndex(currentRow.displayIndex);
		currentRow.displayIndex = i;
	    }
	} else {
	    // undo sorting and stop!
	    unSort();
	    return;
	}
	rowMap = newRowMap;

	sortIndex = colIndex;
	sortMethod = mode;

	// update current row index based on the old one
	if (currentRowIndex >= 0) {
	    currentRowIndex = getRow(oldActualCurrent).displayIndex;
	}

	layout();
	content.repaint();
    }

    /**
     * Unsort the data in the table.
     */
    public void unSort() {
	if (sortMethod != NONE || sortIndex == -1) {
	    if (currentRowIndex >= 0) {
		currentRowIndex = rowMap[currentRowIndex];
	    }
	    sortIndex = -1;
	    sortMethod = NONE;
	    for (int i = 0; i < rowCount; i++) {
		rowMap[i] = i;
		rows[i].displayIndex = i;
	    }
	    layout();
	    content.repaint();
	}
    }

    /**
     * Get the current method of sorting, which can be
     * ascending, descending or none.
     * @see #sortMethod
     * @see #getSortIndex
     */
    public int getSortMethod() {
	return sortMethod;
    }

    /**
     * Get the column index on which currently is being sorted,
     * which is the actual index and not the display index.
     * @see #sortIndex
     * @see #getSortMethod
     */
    public int getSortIndex() {
	return sortIndex;
    }

    /**
     * Set the rows by supplying the array with
     * Vectors containing the data directly.
     * @see #rows
     */
    public void setRows(Vector[] rows) {
	TableRow  newRows[] = new TableRow[rows.length];
	for (int i = 0; i < rows.length; i++) {
	    TableRow  newRow = new TableRow();
	    newRow.data = rows[i];
	    newRows[i] = newRow;
	}
	setRows(newRows);
    }

    /**
     * Set the rows by supplying the array with
     * TableRows directly.
     * @see #rows
     */
    public void setRows(TableRow[] rows) {
	// set the new rows
	removeAllRows();
	this.rows = rows;

	// set the row mapping
	rowMap = new int[rows.length];
	for (int i = 0; i < rows.length; i++) {
	    rowMap[i] = i;
	    rows[i].displayIndex = i;
	}

	// update some other variables
	rowCount = rows.length;

	// make sure sorting is correct
	if (sortIndex >= 0) {
	    sortOnColumnAt(sortIndex, sortMethod);
	}

	layout();
	repaint();
    }

    /**
     * Append an empty row to this table.
     */
    public void addRow() {
	addRow(new Vector());
    }

    /**
     * Append a row to this table.
     * You can supply null to add an empty row.
     */
    public void addRow(Vector rowData) {
	insertRowAt(rowData, rowCount());
    }

    /**
     * Append a row in the form of a TableRow.
     * You can supply null to add an empty row.
     */
    public void addRow(TableRow row) {
	insertRowAt(row, rowCount());
    }

    /**
     * Insert a row at the given position to this table.
     * You can supply null to add an empty row.
     * Its actual index and display index will be the same.
     */
    public void insertRowAt(Vector rowData, int index) {
	insertRowAt(new TableRow("", 0, 0, index, rowData), index);
    }

    /**
     * Insert a row at the given position to this table.
     * You can supply null to add an empty row.
     * Its actual index and display index will be the same.
     */
    public void insertRowAt(TableRow row, int index) {
	// do nothing if one of the parameters is incorrect
	if (index < 0 || index > rowCount) {
	    paramError("insertRowAt(TableRow,int)", "invalid rownumber specified");
	    return;
	}

	// grow arrays if necessary
	if (rowCount + 1 >= rows.length) {
	    TableRow  newRows[] = new TableRow[Math.max(rowCount*2, 4)];
	    System.arraycopy(rows, 0, newRows, 0, rowCount);
	    rows = newRows;

	    int  newRowMap[] = new int[Math.max(rowCount*2, 4)];
	    System.arraycopy(rowMap, 0, newRowMap, 0, rowCount);
	    rowMap = newRowMap;
	}

	// add new row at given position, height and pos are updated by layout()
	System.arraycopy(rows, index, rows, index+1, rowCount-index);
	if (row == null) {
	    row = new TableRow("", 0, 0, index, null);
	} else {
	    row.displayIndex = index;
	}
	rows[index] = row;

	// update row mapping
	System.arraycopy(rowMap, index, rowMap, index+1, rowCount-index);
	rowMap[index] = index;
	for (int i = 0; i < rowCount; i++) {
	    if (i != index && rowMap[i] >= index) {
		rowMap[i]++;
	    }
	}

	// adjust displayIndex of rows displayed below new row
	for (int i = rowCount-1; i > index; i--) {
	    TableRow  currentRow = getRow(getRowIndex(i));
	    currentRow.displayIndex++;
	}

	rowCount++;

	// make sure sorting is correct
	if (sortIndex >= 0) {
	    sortOnColumnAt(sortIndex, sortMethod);
	}

	layout();
	content.repaint();
    }

    /**
     * Remove all the rows.
     */
    public void removeAllRows() {
	rows = new TableRow[4];
	rowCount = 0;
	rowMap = new int[4];
	selectedRows = new int[0];
	layout();
	content.repaint();
    }

    /**
     * Remove the given row.
     */
    public void removeRow(TableRow oldRow) {
	// do nothing if one of the parameters is incorrect
	if (oldRow == null) {
	    paramError("removeRow(row)", "null row specified");
	    return;
	}

	// remove row from selected rows and get the row index
	if (rowIsSelected(oldRow)) {
	    deselectRow(oldRow);
	}
	int  rowIndex = getRowIndex(oldRow.displayIndex);

	// remove row from rows array and prefend memory leak
	System.arraycopy(rows, rowIndex+1, rows, rowIndex, rowCount-rowIndex-1);
	rowCount--;
	rows[rowCount] = null;

	// update row mapping
	if (oldRow.displayIndex != rowCount) {
	    System.arraycopy(rowMap, oldRow.displayIndex+1, rowMap, oldRow.displayIndex, rowCount-oldRow.displayIndex);
	}
	for (int i = 0; i < rowCount; i++) {
	    if (rowMap[i] > rowIndex) {
		rowMap[i]--;
	    }
	}

	// adjust displayIndex of rows below
	for (int i = rowCount-1; i >= oldRow.displayIndex; i--) {
	    TableRow  row = getRow(getRowIndex(i));
	    row.pos -= oldRow.size;
	    row.displayIndex--;
	}

	layout();
	content.repaint();

    }

    /**
     * Remove the row at the given position, which is the
     * actual and not the display index.
     */
    public void removeRowAt(int rowIndex) {
	// do nothing if one of the parameters is incorrect
	if (rowIndex < 0 || rowIndex >= rowCount) {
	    paramError("removeRowAt(int)", "row index " + rowIndex + " does not exist");
	    return;
	}

	removeRow(getRow(rowIndex));
    }

    /**
     * Get the index of the row that is displayed at the given
     * position. Mapping the display index to the actual index is
     * necessary, since you can only search using the actual index.
     * Mapping from actual index to display index can be done by
     * asking for 'displayIndex' in TableRow.
     * @see #rowMap
     */
    public int getRowIndex(int displayIndex) {
	return rowMap[displayIndex];
    }

    /**
     * Select the given row. If the select mode
     * is exclusive it will deselect all other rows.
     * @see #selectedRows
     */
    public void selectRow(TableRow row) {
	// do nothing if one of the parameters is incorrect
	if (row == null) {
	    paramError("selectRow(TableRow)", "null object specified");
	    return;
	}

	selectRowAt(getRowIndex(row.displayIndex));
    }

    /**
     * Select the row with the given (actual) index. If the select
     * mode is exclusive it will deselect all other rows.
     * @see #selectedRows
     */
    public void selectRowAt(int rowIndex) {
	// do nothing if one of the parameters is incorrect
	if (rowIndex < 0 || rowIndex >= rowCount) {
	    paramError("selectRow(int)", "row index "+rowIndex+" out of range");
	    return;
	}

	if (selectMode != MULTIPLE) {
	    deselectAllRows();
	}
	if (selectMode != NONE && !rowIsSelectedAt(rowIndex)) {
	    int[]  newSel = new int[selectedRows.length+1];
	    System.arraycopy(selectedRows, 0, newSel, 0, selectedRows.length);
	    newSel[newSel.length-1] = rowIndex;
	    selectedRows = newSel;

	    content.repaint();
	}
    }

    /**
     * Deselect the given row.
     * @see #selectedRows
     */
    public void deselectRow(TableRow row) {
	// do nothing if one of the parameters is incorrect
	if (row == null) {
	    paramError("deselectRow(TableRow)", "null object specified");
	    return;
	}

	deselectRowAt(getRowIndex(row.displayIndex));
    }

    /**
     * Deselect the row with the given (actual) index.
     * @see #selectedRows
     */
    public void deselectRowAt(int rowIndex) {
	// do nothing if one of the parameters is incorrect
	if (rowIndex < 0 || rowIndex >= rowCount) {
	    paramError("deselectRow(int)", "row index "+rowIndex+" out of range");
	    return;
	}

	// search for the row in the selected rows
	int  selIndex;
	for (selIndex = 0; selIndex < selectedRows.length; selIndex++) {
	    if (selectedRows[selIndex] == rowIndex) {
		break;
	    }
	}

	// deselect if it was selected
	if (selIndex < selectedRows.length) {
	    int[]  newSel = new int[selectedRows.length-1];
	    System.arraycopy(selectedRows, 0, newSel, 0, selIndex);
	    System.arraycopy(selectedRows, selIndex+1, newSel, selIndex, selectedRows.length-selIndex-1);
	    selectedRows = newSel;

	    content.repaint();
	}
    }

    /**
     * Deselect all rows.
     * @see #selectedRows
     */
    public void deselectAllRows() {
	if (selectedRows.length > 0) {
	    selectedRows = new int[0];
	    content.repaint();
	}
    }

    /**
     * Get the currently selected rows as a vector
     * containing TableRow objects.
     * If no row is currently selected, an vector
     * with zero elements is returned.
     * @see #selectedRows
     */
    public Vector getSelectedRows() {
	Vector  result = new Vector();
	for (int i = 0; i < selectedRows.length; i++) {
	    result.addElement(getRow(selectedRows[i]));
	}
	return result;
    }

    /**
     * Get the indexes of the currently selected rows.
     * If no row is currently selected, an array with
     * zero elements is returned.
     * @see #selectedRows
     */
    public int[] getSelectedRowIndexes() {
	// return a copy of the array so it cannot be modified
	int[]  result = new int[selectedRows.length];
	System.arraycopy(selectedRows, 0, result, 0, selectedRows.length);
	return result;
    }

    /**
     * Check whether the given table row is selected.
     * @see #selectedRows
     */
    public boolean rowIsSelected(TableRow row) {
	// do nothing if one of the parameters is incorrect
	if (row == null) {
	    paramError("rowIsSelected(TableRow)", "null object specified");
	    return false;
	}

	return rowIsSelectedAt(getRowIndex(row.displayIndex));
    }

    /**
     * Check whether the row with the given actual index
     * is selected.
     * @see #selectedRows
     */
    public boolean rowIsSelectedAt(int rowIndex) {
	// do nothing if one of the parameters is incorrect
	if (rowIndex < 0 || rowIndex >= rowCount) {
	    paramError("rowIsSelectedAt(int)", "row index "+rowIndex+" out of range");
	    return false;
	}

	boolean  result = false;
	for (int i = 0; i < selectedRows.length; i++) {
	    if (selectedRows[i] == rowIndex) {
		result = true;
		break;
	    }
	}
	return result;
    }

    /**
     * Get the current row, which is the row that has the focus.
     * @see #currentRowIndex
     */
    public TableRow getCurrentRow() {
	return getRow(getRowIndex(currentRowIndex));
    }

    /**
     * Set the current row with the focus to the given row.
     * @see #currentRowIndex
     */
    public void setCurrentRow(TableRow currentRow) {
	if (currentRow == null) {
	    paramError("setCurrentRow(TableRow)","no row specified (null)");
	    return;
	}
	if (currentRowIndex != currentRow.displayIndex) {
	    currentRowIndex = currentRow.displayIndex;
	    content.repaint();
	}
    }

    /**
     * Get the display index of the current row.
     * @see #currentRowIndex
     */
    public int getCurrentRowIndex() {
	return currentRowIndex;
    }

    /**
     * Set the current display row index to the given (display) row index.
     * @see #currentRowIndex
     */
    public void setCurrentRowIndex(int displayRowIndex) {
	if (displayRowIndex < 0 || displayRowIndex >= rowCount) {
	    paramError("setCurrentRowIndex(int)","invalid row index specified");
	    return;
	}
	if (currentRowIndex != displayRowIndex) {
	    currentRowIndex = displayRowIndex;
	    content.repaint();
	}	
    }

    /**
     * When scrolling the content to a given position
     * generate a repaint for the headers-area if necessary.
     */
    public void scrollContent(int tx, int ty) {
	super.scrollContent(tx, ty);
	if (colHeaders != NONE) {
	    repaint(0, 0, width, hdrHeight);
	}
    }

    /**
     * Layout the widget, which means setting the
     * content height and width and laying out the content.
     */
    public void layout() {
	content.layout();
	hdrHeight = (colHeaders == NONE) ? 0 : ((getFontMetrics(hdrFont).getHeight()+4)/5)*5;
	this.contHeight = rowsHeight();
	this.contWidth = columnsWidth();

	// layout the scrolling container
	super.layout();

	// vert. scrollbar must start at the top, not below the headers
	if (vert != null) {
	    vert.reshape(vert.x, 0, vert.width, vert.height + vert.y);
	}
    }

    /**
     * Focus the content at the row with the given actual index.
     */
    public void focusAtRow(int rowIndex) {
	if (rowIndex < 0 || rowIndex >= rowCount) {
	    paramError("focusAtRow(int)","invalid row index specified");
	    return;
	}
	TableRow  row = getRow(rowIndex);
	focusContent(-content.tx, row.pos, 1, row.size + 2);
    }

    /**
     * Paint the headers of the table.
     */
    public void paint(Graphics g, int cx, int cy, int cw, int ch) {
	super.paint(g, cx, cy, cw, ch);

	int  hdrOffset = -content.tx;
	FontMetrics fm = g.getFontMetrics(hdrFont);

	// only paint if there are headers and headers need repainting
	if (colHeaders != NONE && cy < hdrHeight) {
	    // fill if headers must be filled
	    if (fillHeaders) {
		g.setColor(background);
		g.fillRect(0, 0, content.width, hdrHeight);
	    }

	    g.setFont(hdrFont);
	    int  y = fm.getAscent();
	    boolean  lastcolumn = false;
	    int  index;
	    for (index = 0; index < columnCount; index++) {
		TableColumn  col = getColumn(getColumnIndex(index));
		int  colWidth = col.size;

		// column width is smaller when it is partly 'offscreen'
		int  xpos = content.tx + col.pos;
		if (xpos + col.size >= content.width) {
		    colWidth = content.width - col.pos - content.tx;
		}
		if (xpos >= -col.size) {
		    Graphics  g2 = g.create();
		    try {
			g2.clipRect(xpos, 0, colWidth+1, hdrHeight);

			// paint button and check whether this button is down
			boolean  down = this.down && (index == downButton-1);
			if (colHeaders == OUTLINE && !down) {
			    g2.setColor(foreground);
			    g2.drawRect(col.pos - hdrOffset, 0, col.size, hdrHeight+1);
			}
			if (colHeaders == RAISED || down) {
			    g2.setColor(background);
			    Bevel.drawBorder(g2, col.pos - hdrOffset, 0, col.size, hdrHeight+1, !down);
			}

			// paint focus rectangle if this is the button to which is being moved
			if (columnDrag && !down && index == moveColumn) {
			    g.setColor(getFocusColor());
			    Dash.drawFocusRect(g, col.pos - hdrOffset + 2, 2, col.size-4, hdrHeight-3);
			}

			// draw label with a left margin of 5 pixels
			g2.setColor(foreground);
			int  pixelOffset = (down) ? 1 : 0;
			int  lx = col.pos + 5 - hdrOffset + pixelOffset;
			switch (col.align) {
			    case CENTER:
				lx += (col.size - 10 - fm.stringWidth(col.label)) / 2;
				break;
			    case RIGHT:
				lx += (col.size - 10 - fm.stringWidth(col.label));
				break;				
			}
			g2.drawString(col.label, lx, y+pixelOffset);

			// draw sort symbol if appropriate
			if (sortSymbol && getColumnIndex(index) == sortIndex) {
			    int  sortX = col.pos + 15 - hdrOffset + fm.stringWidth(col.label);
			    switch (col.align) {
				case CENTER:
				    sortX += (col.size - 10 - fm.stringWidth(col.label)) / 2;
				    break;
				case RIGHT:
				    sortX = col.pos + col.size - (sortX - col.pos);
				    break;
			    }
			    paintSortSymbol(g2, sortX + pixelOffset, y - 1 + pixelOffset, (hdrHeight - 4) / 2, sortMethod);
			}
		    } finally {
			g2.dispose();
		    }
		}

		// stop when the last column has been painted
		if (lastcolumn) {
		    break;
		}

		// check whether the next column is the last column
		if (xpos + col.size >= content.width) {
		    lastcolumn = true;
		}
	    }

	    // paint an empty column when the last column has been painted
	    if (colHeaders == RAISED || colHeaders == OUTLINE) {
		Graphics  g2 = g.create();
		try {
		    if (index == 0) {
			int  borderWidth = content.width + ((vert == null) ? -1 : 10);
			g2.clipRect(0, 0, content.width, hdrHeight);
			if (colHeaders == RAISED) {
			    g2.setColor(background);
			    Bevel.drawBorder(g2, 0, 0, borderWidth, hdrHeight+1, true);
			} else {
			    g2.setColor(foreground);
			    g2.drawRect(0, 0, borderWidth, hdrHeight+1);
			}
		    } else if (index == columnCount) {
			TableColumn  col = getColumn(getColumnIndex(index-1));
			int  xpos = content.tx + col.pos;
			int  lastWidth = content.width - xpos - col.size;
			int  borderWidth =  lastWidth + ((vert == null) ? -1 : 10);

			g2.clipRect(xpos + col.size, 0, lastWidth + 1, hdrHeight);
			if (colHeaders == RAISED) {
			    g2.setColor(background);
			    Bevel.drawBorder(g2, xpos + col.size, 0, borderWidth, hdrHeight+1, true);
			} else {
			    g2.setColor(foreground);
			    g2.drawRect(xpos + col.size, 0, borderWidth, hdrHeight+1);
			}
		    }
		} finally {
		    g2.dispose();
		}
	    }		
	}
    }

    /**
     * Paint the sort symbol at the given coordinates.
     * @param sorting None, ascending or descending.
     */
    protected void paintSortSymbol(Graphics g, int x, int y, int height, int sorting) {
	int  halfHeight = height / 2 + 1;

	if (sorting == ASCENDING) {
	    g.setColor(background.brighter());
	    g.drawLine(x+halfHeight-1, y-height, x, y);

	    g.setColor(background.darker());
	    g.drawLine(x-1, y, x-halfHeight, y-height);
	    g.drawLine(x-halfHeight, y-height, x+halfHeight-1, y-height);
	} else if (sorting == DESCENDING) {
	    g.setColor(background.brighter());
	    g.drawLine(x-1, y-height, x-halfHeight, y);

	    g.setColor(background.darker());
	    g.drawLine(x+halfHeight-1, y, x, y-height);
	    g.drawLine(x-halfHeight, y, x+halfHeight-1, y);
	}
    }


    /**
     * Returns the display index of the column the given x-coordinate
     * corresponds with. Returns the negative index of an edge if
     * the x-coordinate corresponds with that edge.
     * Returns zero if no edge or column is found.
     */
    protected int findColumn(int x) {
	int  result = 0;
	if (x <= content.width) {
	    int  index = 0;
	    int  hdrOffset = -content.tx;
	    while (index < columnCount && x >= getColumn(getColumnIndex(index)).pos-hdrOffset) {
		TableColumn  col = getColumn(getColumnIndex(index));
		if (x >= col.pos+5-hdrOffset && x <= col.pos+col.size-5-hdrOffset) {
		    result = index+1;
		    break;
		}
		index++;
	    }
	    if (result == 0 && columnCount > 0) {
		// return correct negative number of resize area
		TableColumn  col = getColumn(getColumnIndex(index-1));
		if (x >= col.pos-5-hdrOffset && x <= col.pos+5-hdrOffset) {
		    result = -(index-1);
		} else {
		    // the area at the right of the last column is not a resize area
		    TableColumn  lastCol = getColumn(getColumnIndex(columnCount-1));
		    if (x <= lastCol.pos+lastCol.size+5-hdrOffset) {
			result = -index;
		    }
		}
	    }
	}
	return result;
    }

    /**
     * Checks whether the given event is located in the header area.
     */
    protected boolean eventInHeader(Event evt) {
	TableColumn  lastCol = getColumn(getColumnIndex(columnCount-1));
	int  maxX = Math.min(content.tx + lastCol.pos + lastCol.size + 5, content.width);
	return (evt.y>0 && evt.y<=hdrHeight && evt.x>=0 && evt.x<=maxX);
    }

    /**
     * Event handler for mouse events.
     */
    public boolean handleEvent(Event evt) {
	if (colHeaders != NONE && columnCount > 0) {
	    switch (evt.id) {
		case Event.GOT_FOCUS:
		    content.repaint();
		    break;

		case Event.LOST_FOCUS:
		    content.repaint();
		    break;

		case Event.KEY_PRESS:
		    if (evt.key == '\n' && currentRowIndex >= 0) {
			action(-1, currentRowIndex);
			return true;
		    }
		    break;

		case Event.KEY_ACTION:
		    switch (evt.key) {
			case Event.UP: {
			    deselectAllRows();
			    if (currentRowIndex > 0) {
				currentRowIndex = currentRowIndex - 1;
				focusAtRow(getRowIndex(currentRowIndex));
				selectRowAt(getRowIndex(currentRowIndex));
				postEvent(new Event(this, Event.LIST_SELECT, new Integer(currentRowIndex)));
			    } else {
				selectRowAt(getRowIndex(currentRowIndex));
			    }
			    break;
			}

			case Event.DOWN: {
			    deselectAllRows();
			    if (currentRowIndex < rowCount-1) {
				currentRowIndex = currentRowIndex + 1;
				focusAtRow(getRowIndex(currentRowIndex));
				selectRowAt(getRowIndex(currentRowIndex));
				postEvent(new Event(this, Event.LIST_SELECT, new Integer(currentRowIndex)));
			    } else {
				selectRowAt(getRowIndex(currentRowIndex));
			    }
			    break;
			}
		    }

		case Event.MOUSE_EXIT: {
		    // reset cursor if a column is not being resized
		    if (downButton >= 0) {
			resetCursor();
		    }
		    break;
		}

		case Event.MOUSE_MOVE: {
		    if (!disabled) {
			PlayerPanel p = getPlayerPanel();
			if (p.editing()) {
			    break;
			}
			if (!eventInHeader(evt)) {
			    resetCursor();
			    break;
			}
			if (findColumn(evt.x) < 0 && columnResize) {
			    setCursor(Frame.E_RESIZE_CURSOR);
			} else {
			    resetCursor();
			}
		    }
		    break;
		}

		case Event.MOUSE_DOWN: {
		    if (disabled || !eventInHeader(evt) || evt.target != this) {
			break;
		    }
		    oldX = evt.x;
		    int  colNumber = findColumn(evt.x);
		    if (colNumber > 0) {
			// mouse pointer was in column header
			downButton = colNumber;
			down = true;
			repaint();
		    } else if (colNumber < 0 && columnResize) {
			// mouse pointer was in resize area
			downButton = colNumber;
		    }
		    break;
		}

		case Event.MOUSE_DRAG: {
		    if (downButton > 0) {
			// column header is being pressed
			if (!eventInHeader(evt) || (!columnDrag && (findColumn(evt.x)-1) != downButton-1)) {
			    if (down) {
				down = false;
				moveColumn = -1;
				repaint();
			    }
			    break;
			}
			// event was in header area
			int  colNumber = findColumn(evt.x);
			if (colNumber != downButton) {
			    // check whether switching columns
			    if (colNumber > 0) {
				if (!down || moveColumn != colNumber-1) {
				    moveColumn = colNumber-1;
				    down = true;
				    repaint();
				}
			    } else {
				if (moveColumn >= 0) {
				    moveColumn = -1;
				    repaint();
				}
			    }
			} else {
			    if (!down || moveColumn >= 0) {
				down = true;
				moveColumn = -1;
				repaint();
			    }
			}
		    } else if (downButton < 0 && columnResize) {
			// column is being resized
			setCursor(Frame.E_RESIZE_CURSOR);
			int  colNumber = (-downButton) - 1;
			changeColumnWidth(getColumnIndex(colNumber), evt.x - oldX);
			TableColumn  currentCol = getColumn(getColumnIndex(colNumber));
			oldX = Math.max(evt.x, content.tx + currentCol.pos + currentCol.size);
		    }
		    break;
		}

		case Event.MOUSE_UP: {
		    if (downButton > 0 && down && moveColumn < 0 && sortMode != NONE) {
			//// sort on column ///
			if (getColumnIndex(downButton-1) == sortIndex) {
			    // next sort
			    if (sortMethod == ASCENDING) {
				if (sortMode == BOTH) {
				    sortOnColumnAt(getColumnIndex(downButton-1), DESCENDING);
				} else if (unsort) {
				    unSort();
				}
			    } else if (sortMethod == DESCENDING) {
				if (unsort) {
				    unSort();
				} else if (sortMode == BOTH) {
				    sortOnColumnAt(getColumnIndex(downButton-1), ASCENDING);
				}
			    } else {
				// sortMethod is none
				if (sortMode == DESCENDING) {
				    sortOnColumnAt(getColumnIndex(downButton-1), DESCENDING);
				} else {
				    sortOnColumnAt(getColumnIndex(downButton-1), ASCENDING);
				}
			    }
			} else {
			    // first sort on new column
			    if (sortMode == DESCENDING) {
				sortOnColumnAt(getColumnIndex(downButton-1), DESCENDING);
			    } else {
				sortOnColumnAt(getColumnIndex(downButton-1), ASCENDING);
			    }
			}
		    } else if (columnDrag && downButton > 0 && moveColumn >= 0) {
			//// move column ////
			moveColumn(downButton-1, moveColumn);
		    }

		    if (downButton > 0) {
			downButton = 0;
			moveColumn = -1;
			repaint();
		    } else if (downButton < 0) {
			// column was resized, scroll to make the column entirely visible
			int  colNumber = (-downButton) - 1;
			TableColumn  col = getColumn(getColumnIndex(colNumber));
			int  hdrOffset = -content.tx;
			if (columnsWidth() <= this.width) {
			    focusContent(0, -content.ty, 1, 1);
			} else if (col.pos + col.size < hdrOffset) {
			    focusContent(col.pos + col.size - 1, -content.ty, col.pos + col.size, 1);
			} else if (col.pos + col.size > hdrOffset + this.width) {
			    if (col.size <= this.width) {
				focusContent(col.pos, -content.ty, col.size, 1);
			    } else {
				int  xPos = col.pos + col.size - content.width;
				focusContent(xPos, -content.ty, col.size, 1);
			    }
			}
			layout();
			downButton = 0;
			resetCursor();
		    }
		    break;
		}
	    }
	}
	return super.handleEvent(evt);
    }

    /**
     * The user wants to perform an action at the given cell,
     * represented by the column, row number combination.
     */
    public void action(int column, int row) {
	postEvent(new Event(this, Event.ACTION_EVENT, new Integer(row)));
	action();
    }

    /**
     * The user performed an action.
     */
    public void action() {
	// do nothing here anymore, since action(col,row)
	// already posts an action event.
    }

    /**
     * Debugging.
     */
    public void paramString(StringBuffer buf) {
	super.paramString(buf);
	buf.append(",columnCount=");
	buf.append(columnCount());
	buf.append(",rowCount=");
	buf.append(rowCount());
	buf.append(",columnsWidth=");
	buf.append(columnsWidth());
	buf.append(",rowsHeight=");
	buf.append(rowsHeight());
    }

    /**
     * The table widget supports input focus when it is not
     * disabled and there are rows.
     */
    public boolean focusInterest() {
	return !disabled && rowCount() > 0;
    }

    /**
     * Print error message for incorrect parameters.
     */
    void paramError(String methodName) {
	System.out.println("ERROR: " + methodName + " - incorrect parameters");
    }

    /**
     * Print error message for incorrect parameters.
     */
    void paramError(String methodName, String explanation) {
	System.out.print("ERROR: " + methodName + " - incorrect parameters");
	System.out.println(", "+explanation);
    }

    /**
     * Returns the current column mapping table as a string.
     * DEBUGGING
     */
    String getColMappingString() {
	String  result="display: ";
	for (int index = 0; index < columnCount; index++) {
	    result += String.valueOf(index) + " ";
	}
	result += "\n actual: ";
	for (int index = 0; index < columnCount; index++) {
	    int  actual = getColumnIndex(index);
	    String  actualString = String.valueOf(actual);
	    if (actual == index) {
		actualString = "-";
	    }
	    result += actualString + " ";
	}
	result += "\n";
	return result;
    }

    /**
     * Returns the current row mapping table as a string.
     * DEBUGGING
     */
    String getRowMappingString() {
	String  result="display: ";
	for (int index = 0; index < rowCount; index++) {
	    result += String.valueOf(index) + " ";
	}
	result += "\n actual: ";
	for (int index = 0; index < rowCount; index++) {
	    int  actual = getRowIndex(index);
	    String  actualString = String.valueOf(actual);
	    if (actual == index) {
		actualString = "-";
	    }
	    result += actualString + " ";
	}
	result += "\n";
	return result;
    }

    /**
     * Returns the current columns as a string.
     * DEBUGGING
     */
    String getColString() {
	String result = "";
	for (int index = 0; index < columnCount; index++) {
	    result += getColumn(index).toString() + "\n";
	}
	return result;
    }

    /**
     * Print a debugging message to System.out
     */
    void debugPrint(Object message) {
	System.out.println(message);
    }
}
