// Copyright 1996, Marimba Inc. All Rights Reserved.
// @(#)Widget.java, 1.138, 01/18/97

package marimba.gui;

import java.io.*;
import java.awt.*;
import java.net.URL;
import java.util.Vector;
import java.applet.AudioClip;
import java.awt.image.ImageObserver;
import java.awt.image.ImageProducer;

import marimba.util.QuickSort;
import marimba.util.Timer;
import marimba.util.TimerClient;
import marimba.persist.*;

/**
 * A generic widget implementation. This is the base class for
 * all widgets in the marimba.gui package.
 *
 * @author	Arthur van Hoff
 * @version 	1.138, 01/18/97
 * @see		WidgetConstants
 */
public abstract class Widget implements WidgetConstants, PropertyObject, ImageObserver, TimerClient {
    /**
     * The default font.
     */
    public final static Font defaultFont = new Font("Dialog", Font.PLAIN, 14);

    /**
     * The default foreground color.
     */
    public final static Color defaultForeground = Color.black;

    /**
     * The default background color.
     */
    public final static Color defaultBackground = new Color(192, 192, 192);

    /**
     * The default hilite color,
     * used to fill certain areas to hide the background.
     */
    public final static Color defaultHilite = Color.white;

    /**
     * The name of this widget.
     */
    public String name;

    /**
     * REMIND
     */
    public String wTip;

    /**
     * The tip that is displayed under the mouse pointer
     * when it is pointing at this widget for a while.
     * @see #getTip
     * @see #setTip
     */
    public String tip;

    /**
     * If the widget is invalid, the next time the widget is
     * repainted, the layout has to be redone.
     * @see #validate
     * @see #invalidate
     */
    public boolean valid = true;

    /**
     * If a widget is transparent, the background behind this
     * widget is still visible.
     * @see #isTransparent
     * @see #setTransparent
     */
    protected boolean transparent = true;

    /**
     * If a widget is opaque and so its update method can be called,
     * when directGraphics is true the update method is passed the
     * screen graphics context.  Otherwise the update method draws
     * into a back buffer.  This should be false unless you want to
     * really optimize incremental redisplay yourself.
     * @see #transparent
     */
    protected boolean directGraphics;

    /**
     * A widget can be made invisible.
     * @see #isVisible
     * @see #isShowing
     * @see #show
     * @see #hide
     */
    public boolean visible = true;

    /**
     * A disabled widget does not accept any user input.
     * @see #isDisabled
     * @see #enable
     * @see #disable
     */
    public boolean disabled;

    /**
     * When a widget has focus, a focus rectangle is usually drawn
     * around it telling the user that keyboard input goes to that
     * widget now. A widget can always get the focus.
     * @see #getFocusColor
     * @see #focusInterest
     * @see #currentFocus
     * @see #hasFocus
     * @see #requestFocus
     * @see #clearFocus
     * @see #firstFocus
     * @see #nextFocus
     * @see #previousFocus
     */
    public boolean focus;

    /**
     * The widget is translated in the x-direction
     * using this value.
     * @see #scroll
     * @see #rscroll
     */
    public int tx;

    /**
     * The widget is translated in the y-direction
     * using this value.
     * @see #scroll
     * @see #rscroll
     */
    public int ty;

    /**
     * The x-coordinate of the upper-left corner of this widget.
     */
    public int x;

    /**
     * The y-coordinate of the upper-left corner of this widget.
     */
    public int y;

    /**
     * The width of this widget.
     * @see #resize
     */
    public int width = 100;

    /**
     * The height of this widget.
     * @see #resize
     */
    public int height = 40;

    /**
     * The number of pixels that the distance between the northern
     * side and this widget has to be.
     * @see WidgetConstants#DETACHED
     * @see #setAttach
     */
    public int north = DETACHED;

    /**
     * The number of pixels that the distance between the eatern
     * side and this widget has to be.
     * @see WidgetConstants#DETACHED
     * @see #setAttach
     */
    public int east = DETACHED;

    /**
     * The number of pixels that the distance between the southern
     * side and this widget has to be.
     * @see WidgetConstants#DETACHED
     * @see #setAttach
     */
    public int south = DETACHED;

    /**
     * The number of pixels that the distance between the western
     * side and this widget has to be.
     * @see WidgetConstants#DETACHED
     * @see #setAttach
     */
    public int west = DETACHED;

    /**
     * When wFont is null, font is the inherited value.
     * When wFont is not null, font is the same as wFont, which
     * is the value that overrides the inherited one.
     * @see #font
     */
    public Font wFont;

    /**
     * The current font. The font can be null after construction.
     * In that case getFont() returns wFont.
     * @see #wFont
     * @see #getFont
     * @see #setFont
     */
    public Font font = defaultFont;

    /**
     * When wForeground is null, foreground is the inherited value.
     * When wForeground is not null, foreground is the same as wForeground,
     * which is the value that overrides the inherited one.
     * @see #foreground
     */
    public Color wForeground;

    /**
     * The current foreground. The foreground can be null after construction.
     * In that case getForeground() returns wForeground.
     * @see #wForeground
     * @see #getForeground
     * @see #setForeground
     */
    public Color foreground = defaultForeground;

    /**
     * When wBackground is null, background is the inherited value.
     * When wBackground is not null, background is the same as wBackground,
     * which is the value that overrides the inherited one.
     * @see #background
     */
    public Color wBackground;

    /**
     * The current background. The background can be null after construction.
     * In that case getBackground() returns wBackground.
     * @see #wBackground
     * @see #getBackground
     * @see #setBackground
     */
    public Color background = defaultBackground;

    /**
     * When wHilite is null, hilite is the inherited value.
     * When wHilite is not null, hilite is the same as wHilite,
     * which is the value that overrides the inherited one.
     * @see #hilite
     */
    public Color wHilite;

    /**
     * The current hilite color. The hilite color can be null after construction.
     * In that case getHilite() returns wHilite.
     * When a widget wants to hide its background, it uses this color for certain
     * areas to fill it.
     * @see #wHilite
     * @see #getHilite
     * @see #setHilite
     */
    public Color hilite = defaultHilite;

    /**
     * The parent container of this class.
     */
    public Widget parent;

    /**
     * The number of widgets that are in this widget.
     * @see #widgets
     */
    public int nwidgets;

    /**
     * The widgets that are contained by this widget.
     * @see #nwidgets
     * @see #getWidget
     */
    public Widget widgets[];

    /**
     * The layout manager for this widget.
     * @see #getLayoutMgr
     * @see #setLayoutMgr
     */
    public WidgetLayoutMgr layoutMgr;

    /**
     * The editor for inline editing.
     * @see #getEditor
     * @see #setEditor
     */
    public Object editor;

    /**
     * This is the script which is currently being edited. This script
     * may not reflect the actual state of the widget since it still
     * needs to be compiled
     * @see #script
     */
    public String wScript;

    /**
     * The current script of the widget which reflects the current
     * state of the widget.
     * @see #wScript
     * @see #clazz
     */
    public String script;

    /**
     * The binary representation of the class that constitutes this
     * widget. If the widget does not have a script this variable is
     * null.
     * @see #script
     */
    public byte clazz[];


    /**
     * Construct a widget.
     */
    public Widget() {
    }

    /**
     * Get the properties of this widget.
     */
    public void getProperties(PropertyList list) {
	list.setString("script", wScript, null);
	list.setByteArray("clazz", clazz, null);
	list.setString("name", name, null);
	list.setString("tip", tip, null);
	list.setInteger("x", x, 0);
	list.setInteger("y", y, 0);
	list.setInteger("width", width, 0);
	list.setInteger("height", height, 0);
	list.setInteger("north", north, DETACHED);
	list.setInteger("east", east, DETACHED);
	list.setInteger("south", south, DETACHED);
	list.setInteger("west", west, DETACHED);
	list.setBoolean("visible", visible, true);
	list.setFont("font", wFont, null);
	list.setColor("foreground", wForeground, null);
	list.setColor("background", wBackground, null);
	list.setColor("hilite", wHilite, null);
	list.setObject("layoutmgr", layoutMgr, null);

	getChildProperties(list);
    }

    /**
     * Get the properties for the children of this widget.
     */
    public void getChildProperties(PropertyList list) {
	list.setObjectArray("widgets", nwidgets, widgets);
    }

    /**
     * Set the properties of this widget.
     */
    public void setProperties(PropertyList list) {
	valid = false;
	script = wScript = list.getString("script", null);
	clazz = list.getByteArray("clazz", null);
	name = list.getString("name", null);
	tip = list.getString("tip", null);
	x = list.getInteger("x", 0);
	y = list.getInteger("y", 0);
	width = list.getInteger("width", 0);
	height = list.getInteger("height", 0);
	north = list.getInteger("north", DETACHED);
	east = list.getInteger("east", DETACHED);
	south = list.getInteger("south", DETACHED);
	west = list.getInteger("west", DETACHED);
	visible = list.getBoolean("visible", true);
	wFont = list.getFont("font", null);
	wForeground = list.getColor("foreground", null);
	wBackground = list.getColor("background", null);
	wHilite = list.getColor("hilite", null);
	font = (wFont != null) ? wFont : defaultFont;
	foreground = (wForeground != null) ? wForeground : foreground;
	background = (wBackground != null) ? wBackground : background;
	hilite = (wHilite != null) ? wHilite : hilite;
	layoutMgr = (WidgetLayoutMgr)list.getObject("layoutmgr", null);

	Object objs[] = list.getObjectArray("widgets", null);
	if ((objs == null) || (objs.length == 0)) {
	    nwidgets = 0;
	    widgets = null;
	} else {
	    widgets = new Widget[nwidgets = objs.length];
	    for (int i = nwidgets ; i-- > 0 ;) {
		Widget w = (Widget)objs[i];
		if (w == null) {
		    System.arraycopy(widgets, i+1, widgets, i, --nwidgets - i);
		} else {
		    widgets[i] = w;
		    w.parent = this;
		}
	    }
	}
    }

    /**
     * Apply new properties to an object
     */
    public void applyProperties(PropertyList list) {
	Widget parent = this.parent;
	boolean initalized = getPlayerPanel() != null;
	boolean running = initalized && isShowing();
	if (parent != null) {
	    // Also repaint the resize handles
	    repaintParent(10, x-2, y-2, width+4, height+4);
	}
	if (running) {
	    stopAll();
	}
	if (initalized) {
	    destroyAll();
	}
	repaint();
	setProperties(list);
	if (initalized) {
	    initAll();
	}
	if (running) {
	    startAll();
	}
	invalidate();
	if (parent != null) {
	    parent.invalidate();
	    // Also repaint the resize handles
	    repaintParent(10, x-2, y-2, width+4, height+4);
	}
    }

    /**
     * Return the widget's parent.
     * @see #parent
     */
    public Widget getParent() {
	return parent;
    }

    /**
     * Get the location as a Point object.
     * @see #x
     * @see #y
     */
    public Point location() {
	return new Point(x, y);
    }

    /**
     * Get the size as a Dimension object.
     * @see #width
     * @see #height
     */
    public Dimension size() {
	return new Dimension(width, height);
    }

    /**
     * Get the bounds as a Rectangle object.
     * @see #x
     * @see #y
     * @see #width
     * @see #height
     */
    public Rectangle bounds() {
	return new Rectangle(x, y, width, height);
    }

    /**
     * Get the name.
     * @see #name
     */
    public String getName() {
	return name;
    }

    /**
     * Set the name.
     * @see #name
     */
    public void setName(String name) {
	this.name = name;
    }

    /**
     * Get the tip.
     * @see #tip
     */
    public String getTip() {
	return tip;
    }

    /**
     * Get the tip for a particular x,y location within the widget.
     * @see #tip
     */
    public String getTip(int x, int y) {
	return getTip();
    }

    /**
     * Set the tip.
     * @see #tip
     */
    public void setTip(String tip) {
	this.tip = tip;
	clearPopups();
    }

    /**
     * Get the font.
     * @see #font
     */
    public Font getFont() {
	return (font == null) ? wFont : font;
    }

    /**
     * Set the font.
     * @see #font
     */
    public void setFont(Font f) {
	if ((wFont != f) && ((wFont == null) || !wFont.equals(f))) {
	    wFont = f;
	    if (getPlayerPanel() != null) {
		inheritFont();
		repaint();
	    }
	}
    }

    /**
     * Inherit font.
     * @see #font
     */
    void inheritFont() {
	Font f = (wFont != null) ? wFont : parent.font;
	if (!f.equals(font)) {
	    font = f;
	    invalidate();

	    for (int i = 0 ; i < nwidgets ; i++) {
		Widget w = widgets[i];
		if (w != null) {
		    w.inheritFont();
		}
	    }
	}
    }

    /**
     * Get the foreground color.
     * @see #foreground
     */
    public Color getForeground() {
	return (foreground == null) ? wForeground : foreground;
    }

    /**
     * Set the foreground color.
     * @see #foreground
     */
    public void setForeground(Color c) {
	if ((wForeground != c) && ((wForeground == null) || !wForeground.equals(c))) {
	    wForeground = c;
	    if (getPlayerPanel() != null) {
		inheritForeground();
		repaint();
	    }
	}
    }

    /**
     * Inherit foreground.
     * @see #foreground
     */
    void inheritForeground() {
	Color c = (wForeground != null) ? wForeground : parent.foreground;
	if (!c.equals(foreground)) {
	    foreground = c;
	    for (int i = 0 ; i < nwidgets ; i++) {
		Widget w = widgets[i];
		if (w != null) {
		    w.inheritForeground();
		}
	    }
	}
    }

    /**
     * Get the background color.
     * @see #background
     */
    public Color getBackground() {
	return (background == null) ? wBackground : background;
    }

    /**
     * Set the background color.
     * @see #background
     */
    public void setBackground(Color c) {
	if ((wBackground != c) && ((wBackground == null) || !wBackground.equals(c))) {
	    wBackground = c;
	    if (getPlayerPanel() != null) {
		inheritBackground();
		repaint();
	    }
	}
    }

    /**
     * Inherit background.
     * @see #background
     */
    void inheritBackground() {
	Color c = (wBackground != null) ? wBackground : parent.background;
	if (!c.equals(background)) {
	    background = c;
	    for (int i = 0 ; i < nwidgets ; i++) {
		Widget w = widgets[i];
		if (w != null) {
		    w.inheritBackground();
		}
	    }
	}
    }

    /**
     * Get the hilite color.
     * @see #hilite
     */
    public Color getHilite() {
	return (hilite == null) ? wHilite : hilite;
    }

    /**
     * Set the hilite color.
     * @see #hilite
     */
    public void setHilite(Color c) {
	if ((wHilite != c) && ((wHilite == null) || !wHilite.equals(c))) {
	    wHilite = c;
	    if (getPlayerPanel() != null) {
		inheritHilite();
		repaint();
	    }
	}
    }

    /**
     * Inherit hilite.
     * @see #hilite
     */
    void inheritHilite() {
	Color c = (wHilite != null) ? wHilite : parent.hilite;
	if (!c.equals(hilite)) {
	    hilite = c;
	    for (int i = 0 ; i < nwidgets ; i++) {
		Widget w = widgets[i];
		if (w != null) {
		    w.inheritHilite();
		}
	    }
	}
    }

    /**
     * Check whether this widget is transparent.
     * @see #transparent
     */
    public boolean isTransparent() {
	return transparent;
    }

    /**
     * Set the transparency. Only a widget itself is allowed
     * to do this, since it has to make sure that it actually
     * does paint everything when transparency is set to false.
     * @see #transparent
     */
    protected void setTransparent(boolean transparent) {
	if (this.transparent != transparent) {
	    this.transparent = transparent;
	    repaintParent(0, x, y, width, height);
	}
    }

    /**
     * Get the layout manager for this widget.
     * @see #layoutMgr
     */
    public WidgetLayoutMgr getLayoutMgr() {
	return layoutMgr;
    }

    /**
     * Set the layout manager for this widget.
     * @see #layoutMgr
     */
    public void setLayoutMgr(WidgetLayoutMgr layoutMgr ) {
	this.layoutMgr = layoutMgr;
	invalidate();
    }

    /**
     * Get the current player.
     */
    public PlayerPanel getPlayerPanel() {
	Widget parent = this.parent;
	return (parent != null) ? parent.getPlayerPanel() : null;
    }

    /**
     * Get the presentation.
     */
    public Presentation getPresentation() {
	Widget parent = this.parent;
	return (parent != null) ? parent.getPresentation() : null;
    }

    /**
     * Get the name of the class that can be used for
     * in-place editing. Return null if this widget can't
     * be edited in-place.
     * @see #editor
     */
    public String getEditor() {
	return null;
    }

    /**
     * Set the editor for this widget.
     * @see #editor
     */
    public void setEditor(Object editor) {
	if (this.editor != editor) {
	    this.editor = editor;
	    repaint();
	}
    }

    /**
     * Get the current mouse cursor of the presentation.
     * @see Presentation#cursor
     */
    public int getCursor() {
	Presentation p = getPresentation();
	if (p != null) {
	    return p.getCursor();
	} else {
	    return Frame.DEFAULT_CURSOR;
	}
    }

    /**
     * Set the current mouse cursor in the presentation.
     * @see Presentation#cursor
     */
    public void setCursor(int cursor) {
	Presentation p = getPresentation();
	if (p != null) {
	    p.setCursor(cursor);
	}
    }

    /**
     * Reset the current cursor to the default cursor
     * of the presentation.
     * @see Presentation#defaultCursor
     */
    public void resetCursor() {
	Presentation p = getPresentation();
	if (p != null) {
	    p.resetCursor();
	}
    }

    /**
     * Get font metrics for a font.
     */
    public FontMetrics getFontMetrics(Font font) {
	return Toolkit.getDefaultToolkit().getFontMetrics(font);
    }

    /**
     * Get an image.
     */
    public Image getImage(String src) {
	Presentation p = getPresentation();
	return (p != null) ? p.getImage(src) : null;
    }

    /**
     * Check if disabled.
     * @see #disabled
     */
    public boolean isDisabled() {
	return disabled;
    }

    /**
     * Enable the widget.
     * @see #disabled
     */
    public void enable() {
	disable(false);
    }

    /**
     * Disable the widget.
     * @see #disabled
     */
    public void disable() {
	disable(true);
    }

    /**
     * Disable the widget.
     * @see #disabled
     */
    public void disable(boolean disabled) {
	if (this.disabled != disabled) {
	    this.disabled = disabled;
	    repaint();
	}
    }

    /**
     * Check if visible.
     * @see #visible
     */
    public boolean isVisible() {
	return visible;
    }

    /**
     * Check if showing on screen.
     * @see #visible
     */
    public boolean isShowing() {
	if (visible) {
	    Widget parent = this.parent;
	    return (parent != null) && parent.isShowing();
	}
	return false;
    }

    /**
     * Show
     * @see #visible
     */
    public void show() {
	if (!visible) {
	    Widget parent = this.parent;
	    if (parent != null) {
		parent.invalidate();
	    }

	    visible = true;
	    if (isShowing()) {
		startAll();
	    }
	    repaintParent(0, x, y, width, height);
	}
    }

    /**
     * Show conditionally.
     * @param flag	if true then shows, if false then hides this widget
     * @see #hide
     * @see #show
     */
    public void show(boolean flag) {
	if (flag) {
	    show();
	} else {
	    hide();
	}
    }

    /**
     * Hide this widget.
     * @see #visible
     */
    public void hide() {
	if (visible) {
	    Widget parent = this.parent;
	    if (parent != null) {
		parent.invalidate();
	    }

	    if (isShowing()) {
		stopAll();
	    }

	    repaintParent(0, x, y, width, height);
	    visible = false;
	}
    }

    /**
     * Get the value of the widget. This will return a value object
     * that is "appropriate" for this particular widget type. This will
     * return null in cases where the object does not have a value.
     */
    public Object getValue() {
	return null;
    }

    /**
     * Get the value of a named widget. 
     * @see #getWidget
     */
    public Object getValue(String name) {
	Widget w = getWidget(name);
	return (w == null) ? null : w.getValue();
    }

    /**
     * Set the value of the widget. This only works if the value object
     * is of the appropriate type, or if the value object can be converted
     * to the appropriate type.
     */
    public void setValue(Object value) {
    }

    /**
     * Set the value of a named widget.
     * @see #getWidget
     */
    public void setValue(String name, Object value) {
	Widget w = getWidget(name);
	if (w != null) {
	    w.setValue(value);
	}
    }

    /**
     * Get the text of the widget. This will return the label for objects
     * that have a label. It may return null for objects that don't have
     * a text value or label.
     */
    public String getText() {
	return null;
    }

    /**
     * Get the text of a named widget.
     * @see #getWidget
     */
    public String getText(String name) {
	Widget w = getWidget(name);
	return (w == null) ? null : w.getText();
    }

    /**
     * Set the text of a widget. This will change the label for objects
     * that have a label.
     */
    public void setText(String text) {
    }

    /**
     * Set the text of a named widget.
     * @see #getWidget
     */
    public void setText(String name, String text) {
	Widget w = getWidget(name);
	if (w != null) {
	    w.setText(text);
	}
    }

    /**
     * Goto a page in the current folder.
     * @see FolderWidget#gotoPage
     */
    public void gotoPage(String page) {
	Widget parent = this.parent;
	if (parent != null) {
	    parent.gotoPage(page);
	}
    }

    /**
     * Goto a page in a named folder
     * @see FolderWidget#gotoPage
     * @see #getWidget
     */
    public void gotoPage(String name, String page) {
	Widget w = getWidget(name);
	if (w != null) {
	    w.gotoPage(page);
	}
    }

    /**
     * Check if a point is inside the widget. The point is specified in
     * the parent's coordinate system.
     */
    public boolean inside(int px, int py) {
	return (px >= x) && (px < x + width) && (py >= y) && (py < y + height);
    }

    /**
     * Check if the widget is inside a rectangle. The rectangle is specified in
     * the parent's coordinate system.
     * @param pw	width
     * @param ph	height
     */
    public boolean inside(int px, int py, int pw, int ph) {
	return (px <= x) && (py < y) && (px + pw >= x + width) && (py + ph >= y + height);
    }

    /**
     * Check if the widget overlaps with rectangle. The rectangle is specified in
     * the parent's coordinate system.
     * @param pw	width
     * @param ph	height
     */
    public boolean overlap(int px, int py, int pw, int ph) {
	return (x < px + pw) && (y < py + ph) && (x + width > px) && (y + height > py);
    }

    /**
     * Locate the topmost widget that contains the specified coordinate.
     * The coordinates are specified in the parent's coordinate system.
     * Only visible widgets are considered.
     */
    public Widget locateWidget(int x, int y) {
	x -= this.x;
	y -= this.y;
	if ((x < 0) || (y < 0) || (x >= width) || (y >= height)) {
	    return null;
	}
	x -= tx;
	y -= ty;
	for (int i = nwidgets ; i-- > 0 ;) {
	    Widget w = widgets[i];
	    if ((w != null) && w.visible && w.inside(x, y)) {
		Widget child = w.locateWidget(x, y);
		return (child != null) ? child : w;
	    }
	}
	return this;
    }

    /**
     * Locate all the widgets that intersect with the given point and append them to
     * the vector, so that the front most ends up first.
     * The coordinates are specified in the parent's coordinate system.
     * Both visible and invisible widgets are considered.
     */
    public void locateWidgets(Vector v, int x, int y) {
	x -= this.x + tx;
	y -= this.y + ty;
	for (int i = nwidgets ; i-- > 0 ;) {
	    Widget w = widgets[i];
	    if ((w != null) && w.inside(x, y)) {
		v.addElement(w);
	    }
	}
    }

    /**
     * Locate all the widgets inside a rectangle. The widget are added
     * to the vector so that the front most ends up first.
     * The coordinates are specified in the parent's coordinate system.
     * Both visible and invisible widgets are considered.
     */ 
    public void locateWidgets(Vector v, int x, int y, int width, int height) {
	x -= this.x + tx;
	y -= this.y + ty;
	for (int i = nwidgets ; i-- > 0 ;) {
	    Widget w = widgets[i];
	    if ((w != null) && w.inside(x, y, width, height)) {
		v.addElement(w);
	    }
	}
    }

    /**
     * Paint everyting in the indicated area. The area that
     * needs painting is specified in the parent's coordinates.
     * Calls paint(). Overriding this method in subclasses can
     * break everything, so be sure what you're doing when you
     * do that!
     * @see #paint
     */
    protected void paintAll(Graphics g, int x, int y, int width, int height) {
	if (visible) {
	    if ((x -= this.x) < 0) {
		width += x;
		x = 0;
	    }
	    if (x + width > this.width) {
		width = this.width - x;
	    }
	    if (width > 0) {
		if ((y -= this.y) < 0) {
		    height += y;
		    y = 0;
		}
		if (y + height > this.height) {
		    height = this.height - y;
		}
		if (height > 0) {
		    validate();

		    Graphics gc = g.create(this.x, this.y, this.width, this.height);
		    try {
			gc.translate(tx, ty);
			x -= tx;
			y -= ty;
			try {
			    gc.setFont(font);
			    gc.setColor(foreground);
			    paint(gc, x, y, width, height);
			} catch (ThreadDeath e) {
			    throw e;
			} catch (Throwable e) {
			    e.printStackTrace();
			}
			if (nwidgets > 0) {
			    for (int i = 0 ; i < nwidgets ; i++) {
				Widget w = widgets[i];
				if (w != null) {
				    w.paintAll(gc, x, y, width, height);
				}
			    }
			}
		    } finally {
			gc.dispose();
		    }
		}
	    }
	}
    }

    /**
     * Paint part of the screen.
     */
    public void paint(Graphics g, int x, int y, int width, int height) {
	paint(g);
    }

    /**
     * Paint from scratch. Remember only to paint
     * what is inside the clipping area.
     */
    public void paint(Graphics g) {
    }

    /**
     * Update everyting inside this widget.
     * @see #paint
     */
    void updateAll(Graphics g) {
	Graphics gc = g.create(x, y, width, height);
	try {
	    gc.translate(tx, ty);
	    try {
		gc.setFont(font);
		gc.setColor(foreground);
		update(gc);
	    } catch (ThreadDeath e) {
		throw e;
	    } catch (Throwable e) {
		e.printStackTrace();
	    }
	} finally {
	    gc.dispose();
	}
    }

    /**
     * Update the widget, this will be called in
     * response to a repaint() request. You can
     * assume that the screen is still intact.
     */
    public void update(Graphics g) {
	// Paint the widget itself
	paint(g, -tx, -ty, width, height);

	// Paint the children
	if (nwidgets > 0) {
	    for (int i = 0 ; i < nwidgets ; i++) {
		Widget w = widgets[i];
		if (w != null) {
		    w.paintAll(g, -tx, -ty, width, height);
		}
	    }
	}
    }

    /**
     * Repaint the entrire widget as soon as possible.
     */
    public void repaint() {
	repaint(0, -tx, -ty, width, height);
    }

    /**
     * Repaint this widget. If the widget is not obscured
     * and if the wiget is not transparent this will result
     * in a call to update().
     */
    public void repaint(long tm) {
	repaint(tm, -tx, -ty, width, height);
    }

    /**
     * This will repaint the indicated area of
     * the screen (local coordinates). Note that
     * this will always result in a call to paint().
     */
    public void repaint(int x, int y, int width, int height) {
	repaint(0, x, y, width, height);
    }

    /**
     * Repaint a part of the parent window. This is used when reshaping
     * widgets, or when repainting transparent widgets. The coordinates
     * are in the parent's coordinate system.
     */
    public void repaintParent(long tm, int x, int y, int width, int height) {
	Widget parent = this.parent;
	if (visible && (parent != null)) {
	    parent.repaint(tm, x, y, width, height);
	}
    }

    /**
     * This will repaint the indicated area of
     * the screen (local coordinates). Note that
     * this will always result in a call to paint().
     */
    public void repaint(long tm, int x, int y, int width, int height) {
	//System.out.println("rep: " + tm + "," + x + "," + y + "," + width + "x" + height + " - " + getWidgetClass().getName());

	Widget parent = this.parent;
	if (visible && (parent != null)) {
	    if ((x += tx) < 0) {
		width += x;
		x = 0;
	    }
	    if (x + width > this.width) {
		width = this.width - x;
	    }
	    if (width > 0) {
		if ((y += ty) < 0) {
		    height += y;
		    y = 0;
		}
		if (y + height > this.height) {
		    height = this.height - y;
		}
		if (height > 0) {
		    if (transparent) {
			parent.repaint(tm, this.x + x, this.y + y, width, height);
		    } else {
			parent.repaint(tm, this, this, this.x + x, this.y + y, width, height);
		    }
		}
	    }
	}
    }

    /**
     * Repaint an opague widget. 
     */
    void repaint(long tm, Widget w, Widget c, int x, int y, int width, int height) {
	Widget parent = this.parent;
	if (visible && (parent != null)) {
	    if ((x += tx) < 0) {
		width += x;
		x = 0;
	    }
	    if (x + width > this.width) {
		width = this.width - x;
	    }
	    if (width > 0) {
		if ((y += ty) < 0) {
		    height += y;
		    y = 0;
		}
		if (y + height > this.height) {
		    height = this.height - y;
		}
		if (height > 0) {
		    int i = 0;
		    while ((i < nwidgets) && (widgets[i++] != c));
		    while (i < nwidgets) {
			Widget ww = widgets[i++];
			if (ww.visible && ww.overlap(x, y, width, height)) {
			    parent.repaint(tm, this.x + x, this.y + y, width, height);
			    return;
			}
		    }
		    parent.repaint(tm, w, this, this.x + x, this.y + y, width, height);
		}
	    }
	}
    }

    /**
     * Update the widget after more image data arrives.
     */
    public boolean imageUpdate(Image img, int flags, int x, int y, int width, int height) {
	if ((flags & ALLBITS) != 0) {
	    repaint();
	}
	return (flags & (ALLBITS|ERROR|ABORT)) == 0;
    }

    /**
     * Creates an image from the specified image producer.
     * @param producer the image producer
     */
    public Image createImage(ImageProducer producer) {
	return getPlayerPanel().createImage(producer);
    }

    /**
     * Creates an off-screen drawable Image to be used for double buffering.
     * @param width the specified width
     * @param height the specified height
     */
    public Image createImage(int width, int height) {
	return getPlayerPanel().createImage(width, height);
    }

    /**
     * The clock is ticking...
     * @see TimerClient
     */
    public long tick(long tm, Object arg) {
	return -1;
    }

    /**
     * Layout the widget.
     */
    public void layout() {
	if (layoutMgr != null) {
	    layoutMgr.layout(this);
	    return;
	}

	for (int i = 0 ; i < nwidgets ; i++) {
	    Widget w = widgets[i];

	    int x1 = w.x;
	    int y1 = w.y;
	    int x2 = x1 + w.width;
	    int y2 = y1 + w.height;

	    if (w.west < DETACHED) {
		if (w.east < DETACHED) {
		    // stretch
		    x1 = w.west;
		    x2 = width - w.east;
		} else {
		    // move
		    x1 = w.west;
		    x2 = x1 + w.width;
		}
	    } else if (w.east < DETACHED) {
		// move
		x2 = width - w.east;
		x1 = x2 - w.width;
	    }
	    if (w.north < DETACHED) {
		if (w.south < DETACHED) {
		    // stretch
		    y1 = w.north;
		    y2 = height - w.south;
		} else {
		    // move
		    y1 = w.north;
		    y2 = y1 + w.height;
		}
	    } else if (w.south < DETACHED) {
		// move
		y2 = height - w.south;
		y1 = y2 - w.height;
	    }
	    w.reshape(x1, y1, x2-x1, y2-y1);
	}
    }

    /**
     * Validate the widget.
     * @see #valid
     */
    public void validate() {
	if (visible && !valid) {
	    valid = true;
	    layout();
	    valid = true;
	}
    }

    /**
     * Invalidate the widget (the size has changed).
     * @see #valid
     */
    public void invalidate() {
	valid = false;
    }

    /**
     * Reshape the widget.
     * The coordinates are specified in the parent's coordinate system.
     */
    public void reshape(int x, int y, int width, int height) {
	reshape(x, y, width, height, true);
    }

    /**
     * Reshape the widget.
     * The coordinates are specified in the parent's coordinate system.
     */
    public void reshape(int x, int y, int width, int height, boolean repaint) {
	if (width < 0) {
	    width = 0;
	}
	if (height < 0) {
	    height = 0;
	}

	boolean sizeChange = (this.width != width) || (this.height != height);
	if (sizeChange || (this.x != x) || (this.y != y)) {
	    int x1 = this.x, y1 = this.y;
	    int x2 = x1 + this.width, y2 = y1 + this.height;

	    this.x = x;
	    this.y = y;
	    if (sizeChange) {
		this.width = width;
		this.height = height;
		if ((width > 0) && (height > 0)) {
		    invalidate();
		}
	    }

	    Widget parent = this.parent;
	    if (parent != null) {
		parent.invalidate();

 		if (repaint) {
		    if (x1 > x) {
			x1 = x;
		    }
		    if (y1 > y) {
			y1 = y;
		    }
		    if (x2 < x + width) {
			x2 = x + width;
		    }
		    if (y2 < y + height) {
			y2 = y + height;
		    }
		    repaintParent(10, x1, y1, x2-x1, y2-y1);
		}

		if (north < DETACHED) {
		    north = y;
		}
		if (east < DETACHED) {
		    east = parent.width - (x + width);
		}
		if (south < DETACHED) {
		    south = parent.height - (y + height);
		}
		if (west < DETACHED) {
		    west = x;
		}
	    }
	}
    }

    /**
     * Move the widget.
     * The coordinates are specified in the parent's coordinate system.
     * @see #reshape
     */
    public void move(int x, int y) {
	reshape(x, y, width, height);
    }

    /**
     * Move relative.
     * @see #reshape
     */
    public void rmove(int dx, int dy) {
	reshape(x + dx, y + dy, width, height);
    }

    /**
     * Resize the widget.
     */
    public void resize(Dimension d) {
	reshape(x, y, d.width, d.height);
    }

    /**
     * Resize the widget.
     */
    public void resize(int width, int height) {
	reshape(x, y, width, height);
    }

    /**
     * Set the attachments of this widget to the given
     * number of pixels. Use the constant DETACHED from
     * widgetConstants to detach the widget from that side.
     */
    public void setAttach(int north, int east, int south, int west) {
	this.north = (north >= 0) ? north : 0;
	this.east = (east >= 0) ? east : 0;
	this.south = (south >= 0) ? south : 0;
	this.west = (west >= 0) ? west : 0;
    }

    /**
     * Scroll the contents of this widget. This will
     * change the translation of the widget.
     * @see #tx
     * @see #ty
     */
    public void scroll(int tx, int ty) {
	if ((tx != this.tx) || (ty != this.ty)) {
	    this.tx = tx;
	    this.ty = ty;
	    repaint();
	}
    }

    /**
     * Scroll the contents of this widget relative. This will
     * change the translation of the widget.
     * @see #tx
     * @see #ty
     */
    public void rscroll(int dx, int dy) {
	scroll(tx + dx, ty + dy);
    }

    /**
     * Add a widget.
     * @see #widgets
     * @see #nwidgets
     */
    public void add(Widget w) {
	add(nwidgets, w);
    }

    /**
     * Add a widget at a given position.
     * @see #widgets
     * @see #nwidgets
     */
    public void add(int pos, Widget w) {
	if (w != null) {
	    if (w.parent != null) {
		w.parent.remove(w);
	    }
	    if (pos < 0) {
		pos = 0;
	    } else if (pos > nwidgets) {
		pos = nwidgets;
	    }
	    if (widgets == null) {
		widgets = new Widget[4];
	    } else if (nwidgets == widgets.length) {
		Widget newwidgets[] = new Widget[nwidgets*2];
		System.arraycopy(widgets, 0, newwidgets, 0, nwidgets);
		widgets = newwidgets;
	    }
	    System.arraycopy(widgets, pos, widgets, pos + 1, nwidgets++ - pos);
	    
	    widgets[pos] = w;
	    invalidate();
	    repaint(0, w.x, w.y, w.width, w.height);
	    w.parent = this;

	    if (getPlayerPanel() != null) {
		w.initAll();
		if (w.isShowing()) {
		    w.startAll();
		}
	    }
	}
    }

    /**
     * Replace this widget with another widget. This is usually
     * used when scripting widgets.
     */
    public void replace(Widget newWidget) {
	Widget parent = this.parent;
	if (parent != null) {
	    parent.replace(this, newWidget);
	}
    }
    
    /**
     * Replace a widget by another widget.
     */
    public void replace(Widget oldWidget, Widget newWidget) {
	for (int i = 0 ; i < nwidgets ; i++) {
	    if (widgets[i] == oldWidget) {
		replace(i, newWidget);
		return;
	    }
	}
    }

    /**
     * Replace a widget at a given position.
     */
    public void replace(int pos, Widget w) {
	if (w != null) {
	    boolean  initialize = getPlayerPanel() != null;
	
	    if (pos < 0) {
		pos = 0;
	    } else if (pos >= nwidgets) {
		pos = nwidgets-1;
	    }
	    Widget old = widgets[pos];
	    if (initialize) {
		if (old.isShowing()) {
		    old.stopAll();
		}
		old.destroyAll();
	    }
	    old.parent = null;
	    widgets[pos] = null;
	    repaint(old.x, old.y, old.width, old.height);

	    if (w.parent != null) {
		w.parent.remove(w);
	    }
	    widgets[pos] = w;
	    w.parent = this;
	    invalidate();
	    repaint(w.x, w.y, w.width, w.height);
	    if (initialize) {
		w.initAll();
		if (w.isShowing()) {
		    w.startAll();
		}
	    }
	}
    }

    /**
     * Remove this widget from its container.
     * @see #widgets
     * @see #nwidgets
     */
    public void remove() {
	Widget parent = this.parent;
	if (parent != null) {
	    parent.remove(this);
	}
    }
	
    /**
     * Remove a widget from this widget.
     * @see #widgets
     * @see #nwidgets
     */
    public void remove(Widget w) {
	if (w != null) {
	    for (int i = 0 ; i < nwidgets ; i++) {
		if (widgets[i] == w) {
		    remove(w, i);
		    invalidate();
		    repaint(w.x, w.y, w.width, w.height);
		}
	    }
	}
    }

    /**
     * Removes the widget, assuming that the given index
     * is the index of the widget in the widgets array.
     * It does not invalidate and/or repaint this widget.
     */
    protected void remove(Widget w, int i) {
	if (getPlayerPanel() != null) {
	    if (w.isShowing()) {
		w.stopAll();
	    }
	    w.destroyAll();
	}
	w.parent = null;
	System.arraycopy(widgets, i+1, widgets, i, --nwidgets - i);
	if (nwidgets == 0) {
	    widgets = null;
	} else {
	    widgets[nwidgets] = null;
	}
    }


    /**
     * Remove all widgets from this widget.
     */
    public synchronized void removeAll() {
	int  oldCount = nwidgets;

	// variables for area that needs to be updated afterwards
	int  x1 = this.width;
	int  y1 = this.height;
	int  x2 = 0, y2 = 0;
	for (int i = 0; i < nwidgets; i++) {
	    Widget  w = widgets[i];
	    x1 = Math.min(x1, w.x);
	    y1 = Math.min(y1, w.y);
	    x2 = Math.max(x2, w.x + w.width);
	    y2 = Math.max(y2, w.y + w.height);

	    if (getPlayerPanel() != null) {
		if (w.isShowing()) {
		    w.stopAll();
		}
		w.destroyAll();
	    }
	    w.parent = null;
	}
	widgets = null;
	nwidgets = 0;

	if (oldCount != nwidgets) {
	    invalidate();
	    repaint(x1, y1, x2 - x1, y2 - y1);
	}
    }

    /**
     * Bring this widget to the front.
     */
    public void toFront() {
	Widget parent = this.parent;
	if (parent != null) {
	    parent.toFront(this);
	}
    }

    /**
     * Move a widget to the front.
     */
    public void toFront(Widget w) {
	for (int i = 0 ; i < nwidgets ; i++) {
	    if (widgets[i] == w) {
		System.arraycopy(widgets, i+1, widgets, i, nwidgets-(i+1));
		widgets[nwidgets-1] = w;
		invalidate();
		repaint(0, w.x, w.y, w.width, w.height);
		break;
	    }
	}
    }

    /**
     * Send this widget to the back.
     */
    public void toBack() {
	Widget parent = this.parent;
	if (parent != null) {
	    parent.toBack(this);
	}
    }

    /**
     * Move a widget to the back.
     */
    public void toBack(Widget w) {
	for (int i = 0 ; i < nwidgets ; i++) {
	    if (widgets[i] == w) {
		System.arraycopy(widgets, 0, widgets, 1, i);
		widgets[0] = w;
		invalidate();
		repaint(0, w.x, w.y, w.width, w.height);
		break;
	    }
	}
    }

    /**
     * Called to initialize a widget.
     */
    public void init() {
    }

    /**
     * Initialize this widget and all of its children.
     * @see #init
     */
    void initAll() {
	// Inherit values
	font = (wFont != null) ? wFont : parent.font;
	foreground = (wForeground != null) ? wForeground : parent.foreground;
	background = (wBackground != null) ? wBackground : parent.background;
	hilite = (wHilite != null) ? wHilite : parent.hilite;

	try {
	    init();
	} catch (ThreadDeath e) {
	    throw e;
	} catch (Throwable e) {
	    e.printStackTrace();
	}
	for (int i = 0 ; i < nwidgets ; i++) {
	    Widget w = widgets[i];
	    if (w != null) {
		w.initAll();
	    }
	}
    }

    /**
     * Called to start a widget.
     */
    public void start() {
    }

    /**
     * Start this widget and all of its children.
     * @see #start
     */
    void startAll() {
	if (visible) {
	    try {
		start();
	    } catch (ThreadDeath e) {
		throw e;
	    } catch (Throwable e) {
		e.printStackTrace();
	    }
	    for (int i = 0 ; i < nwidgets ; i++) {
		Widget w = widgets[i];
		if (w != null) {
		    w.startAll();
		}
	    }
	}
    }

    /**
     * Called to stop a widget.
     */
    public void stop() {
    }

    /**
     * Stop this widget and all of its children.
     * @see #stop
     */
    void stopAll() {
	if (visible) {
	    for (int i = 0 ; i < nwidgets ; i++) {
		Widget w = widgets[i];
		if (w != null) {
		    w.stopAll();
		}
	    }
	    if (wTip != null) {
		wTip = null;
		clearPopups();
	    }
	    try {
		stop();
	    } catch (ThreadDeath e) {
		throw e;
	    } catch (Throwable e) {
		e.printStackTrace();
	    }
	}
    }

    /**
     * Called to destroy a widget.
     */
    public void destroy() {
	//System.out.println("DESTROY:" + this);
    }

    /**
     * Destroy this widget and all of its children.
     * @see #destroy
     */
    void destroyAll() {
	for (int i = 0 ; i < nwidgets ; i++) {
	    Widget w = widgets[i];
	    if (w != null) {
		w.destroyAll();
	    }
	}
	try {
	    destroy();
	} catch (ThreadDeath e) {
	    throw e;
	} catch (Throwable e) {
	    e.printStackTrace();
	}
    }

    /**
     * Find a widget in this container. This method will only
     * look in the current container. It does not accept hierachical
     * names.
     * @see #getWidget
     */
    public Widget findWidget(String nm) {
	for (int i = 0 ; i < nwidgets ; i++) {
	    Widget w = widgets[i];
	    if (nm.equals(w.name)) {
		return w;
	    }
	}
	if (nm.equals("parent")) {
	    return parent;
	}
	return null;
    }

    /**
     * Get a widget by name. This methods accepts hierarchical
     * names. If the root name is not found in this container
     * the parent of the container is tried.
     * @see #findWidget
     */
    public Widget getWidget(String name) {
	int i = name.indexOf('.');
	String nm = (i < 0) ? name : name.substring(0, i);

	// Look for a widget by this name, up in the hierarchy
	for (Widget p = this ; p != null ; p = p.parent) {
	    Widget w = p.findWidget(nm);
	    if (w != null) {
		while ((w != null) && (i >= 0)) {
		    int j = name.indexOf('.', i+1);
		    w = w.findWidget((j < 0) ? name.substring(i+1) : name.substring(i+1, j));
		    i = j;
		}
		return w;
	    }
	}
	return null;
    }

    /**
     * Get an audio clip.
     */
    public AudioClip getAudioClip(String src) {
	Presentation p = getPresentation();
	return (p != null) ? p.getAudioClip(src) : null;
    }

    /**
     * Get an audio clip using a url.
     */
    public AudioClip getAudioClipAt(URL url) {
	Presentation p = getPresentation();
	return (p != null) ? p.getAudioClipAt(url) : null;
    }

    /**
     * THIS METHOD WILL BE REMOVED IN the beta4 release,
     * It is still here for backward compatibility.
     * New method: audioPlay.
     * @see #audioPlay
     */
    public void play(String src) {
	audioPlay(src);
    }

    /**
     * Play an audio clip.
     */
    public void audioPlay(String src) {
	AudioClip clip = getAudioClip(src);
	if (clip != null) {
	    clip.play();
	}
    }

    /**
     * Play an audio clip using a url.
     */
    public void audioPlay(URL url) {
	AudioClip  clip = getAudioClipAt(url);
	if (clip != null) {
	    clip.play();
	}
    }

    /**
     * Play an audio clip looped.
     */
    public void audioLoop(String src) {
	AudioClip clip = getAudioClip(src);
	if (clip != null) {
	    clip.loop();
	}
    }

    /**
     * Play the audio clip at the given url looped.
     */
    public void audioLoop(URL url) {
	AudioClip clip = getAudioClipAt(url);
	if (clip != null) {
	    clip.loop();
	}
    }

    /**
     * Stop an audio clip.
     */
    public void audioStop(String src) {
	AudioClip clip = getAudioClip(src);
	if (clip != null) {
	    clip.stop();
	}	
    }

    /**
     * Stop the audio clip with the given url.
     */
    public void audioStop(URL url) {
	AudioClip clip = getAudioClipAt(url);
	if (clip != null) {
	    clip.stop();
	}	
    }

    /**
     * Get the selection foreground color.
     * @see #getSelBackground
     */
    public Color getSelForeground() {
	Presentation presentation = getPresentation();
	return (presentation != null) ? presentation.getSelForeground() : Color.white;
    }

    /**
     * Get the selection background color.
     * @see #getSelForeground
     */
    public Color getSelBackground() {
	Presentation presentation = getPresentation();
	return (presentation != null) ? presentation.getSelBackground() : new Color(0, 0, 124);
    }

    /**
     * Get the selection focus color,
     * which is the color of the focus rectangle around the selection bar.
     * @see #getFocusColor
     */
    public Color getSelFocusColor() {
	if (!(getSelBackground().equals(Color.yellow))) {
	    return Color.yellow;
	} else {
	    return Color.black;
	}
    }

    /**
     * Get the focus color,
     * which is the standard color of the focus rectangle.
     * @see #getSelFocusColor
     */
    public Color getFocusColor() {
	if (!(getPresentation().getBackground().equals(Color.black))) {
	    return Color.black;
	} else {
	    return Color.lightGray;
	}
    }

    /**
     * Get the default button.
     * Returns null if no default button is found or if no playerpanel exists.
     * Returns the first default button found, if more default buttons exist.
     */
    public Widget getDefaultButton() {
	PlayerPanel p = getPlayerPanel();
	return (p!=null) ? p.getDefaultButton() : null;
    }

    /**
     * True if this widget is potentially interested in getting the focus.
     * When this function returns false, this does not mean it cannot get
     * the focus. It just means that you cannot give it the focus using
     * the keyboard, you have to do that programmatically.
     * @see #focus
     * @see #requestFocus
     */
    public boolean focusInterest() {
	return false;
    }

    /**
     * Get the widget that has the current focus.
     * @see #focus
     */
    public Widget currentFocus() {
	PlayerPanel p = getPlayerPanel();
	return (p != null) ? p.currentFocus() : null;
    }

    /**
     * True if the current widget has the focus.
     * @see #focus
     */
    public boolean hasFocus() {
	return focus;
    }

    /**
     * Ask the playerPanel for the focus. When the playerPanel has
     * given the focus to this widget, it sends a GOT_FOCUS event.
     * A widget can always get the focus using this function,
     * even when it is not interested in the focus.
     * @see #focus
     * @see #focusInterest
     */
    public void requestFocus() {
	PlayerPanel p = getPlayerPanel();
	if (p != null) {
	    p.requestFocus(this);
	}
    }

    /**
     * Remove the character focus.
     * @see #focus
     */
    public void clearFocus() {
	PlayerPanel player = getPlayerPanel();
	if (player != null) {
	    player.requestFocus(null);
	}
    }

    /**
     * Assign the focus to the first widget.
     * @see #focus
     */
    public void firstFocus() {
	PlayerPanel player;
	Presentation p;

	if (((player = getPlayerPanel()) != null) && ((p = player.getPresentation()) != null)) {
	    // Create an array of focusable widgets
	    Vector v = new Vector();
	    p.getFocusCandidates(v, 0, 0);
	    int n = v.size();
	    if (n > 0) {
		FocusCandidate ws[] = new FocusCandidate[n];
		v.copyInto(ws);

		// Sort the array based on x,y location
	        QuickSort.sort(ws, 0, ws.length, null);

		// Pick the first one
		Widget focus = ws[0].widget;
		if (focus instanceof CommandButtonWidget) {
		    Widget def = getDefaultButton();
		    if (def != null) {
			focus = def;
		    }
		}
		focus.requestFocus();
	    } else {
		clearFocus();
	    }
	}
    }

    /**
     * Assign the focus to the next widget that is located
     * to the right or below the current focus.
     * @see #focus
     */
    public void nextFocus() {
	findFocus(true);
    }

    /**
     * Assign the focus to the previous widget that is located
     * to the left or above the current focus.
     * @see #focus
     */
    public void previousFocus() {
	findFocus(false);
    }

    /**
     * Assign the focus to the previous or next widget.
     * @see #focus
     */
    void findFocus(boolean forward) {
	PlayerPanel player;
	Presentation p;

	if (((player = getPlayerPanel()) != null) && ((p = player.getPresentation()) != null)) {
	    // Create an array of focusable widgets
	    Vector v = new Vector();
	    p.getFocusCandidates(v, 0, 0);
	    int n = v.size();
	    if (n > 0) {
		FocusCandidate ws[] = new FocusCandidate[n];
		v.copyInto(ws);

		// Sort the array based on x,y location
		QuickSort.sort(ws, 0, n, null);

		// Pick the correct candidate
		Widget w = currentFocus();
		if (forward) {
		    int i = 0;
		    if (w != null) {
			while ((i < ws.length) && (ws[i++].widget != w));
		    }
		    if (i < ws.length) {
			ws[i].widget.requestFocus();
		    } else {
			ws[0].widget.requestFocus();
		    }
		} else {
		    int i = ws.length - 1;
		    if (w != null) {
			while ((i >= 0) && (ws[i--].widget != w));
		    }
		    if (i >= 0) {
			ws[i].widget.requestFocus();
		    } else {
			ws[ws.length - 1].widget.requestFocus();
		    }
		}
	    }
	}
    }

    /**
     * Gather all the focusable widgets into a vector.
     * @see #focus
     */
    void getFocusCandidates(Vector v, int x, int y) {
	x -= tx;
	y -= ty;
	for (int i = 0 ; i < nwidgets ; i++) {
	    Widget w = widgets[i];
	    if (w.visible) {
		if (w.focusInterest()) {
		    v.addElement(new FocusCandidate(w, x, y));
		}
		w.getFocusCandidates(v, x + w.x, y + w.y);
	    }
	}
    }

    /**
     * Clear any popups associated with this widget.
     */
    public void clearPopups() {
	PlayerPanel player = getPlayerPanel();
	if (player != null) {
	    player.clearPopups(this);
	}
    }

    /**
     * Post an event to this widget. If the widget does not
     * handle the event, it is passed to the parent.
     * @see #handleEvent
     */
    public boolean postEvent(Event evt) {
	Widget parent = this.parent;
	if (handleEvent(evt)) {
	    return true;
	}

	if (parent != null) {
	    evt.translate(x + tx, y + ty);
	    return parent.postEvent(evt);
	}
	return false;
    }

    /**
     * Handle an event.
     * @see #postEvent
     */
    public boolean handleEvent(Event evt) {
	switch (evt.id) {
	  case Event.MOUSE_ENTER:
	    return mouseEnter(evt, evt.x, evt.y);

	  case Event.MOUSE_EXIT:
	    return mouseExit(evt, evt.x, evt.y);

	  case Event.MOUSE_MOVE:
	    return mouseMove(evt, evt.x, evt.y);

	  case Event.MOUSE_DOWN:
	    return mouseDown(evt, evt.x, evt.y);

	  case Event.MOUSE_DRAG:
	    return mouseDrag(evt, evt.x, evt.y);

	  case Event.MOUSE_UP:
	    return mouseUp(evt, evt.x, evt.y);

	  case Event.KEY_PRESS:
	    switch (evt.key) {
	      case '\t':
		if (evt.controlDown()) {
		    return false;
		}
		if (evt.shiftDown()) {
		    previousFocus();
		} else {
		    nextFocus();
		}
		return true;

	      case '\n':
		Widget  defaultButton = getDefaultButton();
		if (defaultButton != null) {
		    defaultButton.requestFocus();
		    defaultButton.action();
		}
		return true;
	    }
	  case Event.KEY_ACTION:
	    return keyDown(evt, evt.key);

	  case Event.KEY_RELEASE:
	  case Event.KEY_ACTION_RELEASE:
	    return keyUp(evt, evt.key);
	    
	  case Event.ACTION_EVENT:
	    return action(evt, evt.arg);
	  case Event.GOT_FOCUS:
	    repaint();
	    return gotFocus(evt, evt.arg);
	  case Event.LOST_FOCUS:
	    repaint();
	    return lostFocus(evt, evt.arg);

	  case REST_EVENT: {
	    if ((wTip = getTip(evt.x, evt.y)) != null) {
		new Tip(this, wTip, evt.x, evt.y);
		return true;
	    }
	    break;
	  }

	  case WAKE_EVENT:
	    if (wTip != null) {
		wTip = null;
		clearPopups();
		return true;
	    }
	    break;
	}
	return false;
    }

    /**
     * By default a widget calls this function when a
     * mouseDown event occurs.
     * We strongly advise against the use of this function
     * and advise the use of doing the actions within handleEvent
     * to reduce the number of method calls.
     */
    public boolean mouseDown(Event evt, int x, int y) {
	return false;
    }

    /**
     * By default a widget calls this function when a
     * mouseDrag event occurs.
     * We strongly advise against the use of this function
     * and advise the use of doing the actions within handleEvent
     * to reduce the number of method calls.
     */
    public boolean mouseDrag(Event evt, int x, int y) {
	return false;
    }

    /**
     * By default a widget calls this function when a
     * mouseUp event occurs.
     * We strongly advise against the use of this function
     * and advise the use of doing the actions within handleEvent
     * to reduce the number of method calls.
     */
    public boolean mouseUp(Event evt, int x, int y) {
	return false;
    }

    /**
     * By default a widget calls this function when a
     * mouseMove event occurs.
     * We strongly advise against the use of this function
     * and advise the use of doing the actions within handleEvent
     * to reduce the number of method calls.
     */
    public boolean mouseMove(Event evt, int x, int y) {
	return false;
    }

    /**
     * By default a widget calls this function when a
     * mouseEnter event occurs.
     * We strongly advise against the use of this function
     * and advise the use of doing the actions within handleEvent
     * to reduce the number of method calls.
     */
    public boolean mouseEnter(Event evt, int x, int y) {
	return false;
    }

    /**
     * By default a widget calls this function when a
     * mouseExit event occurs.
     * We strongly advise against the use of this function
     * and advise the use of doing the actions within handleEvent
     * to reduce the number of method calls.
     */
    public boolean mouseExit(Event evt, int x, int y) {
	return false;
    }

    /**
     * By default a widget calls this function when a
     * keyDown event occurs.
     * We strongly advise against the use of this function
     * and advise the use of doing the actions within handleEvent
     * to reduce the number of method calls.
     */
    public boolean keyDown(Event evt, int key) {
	return false;
    }

    /**
     * By default a widget calls this function when a
     * keyUp event occurs.
     * We strongly advise against the use of this function
     * and advise the use of doing the actions within handleEvent
     * to reduce the number of method calls.
     */
    public boolean keyUp(Event evt, int key) {
	return false;
    }

    /**
     * By default a widget calls this function when a
     * action event occurs.
     * We strongly advise against the use of this function
     * and advise the use of doing the actions within handleEvent
     * to reduce the number of method calls.
     */
    public boolean action(Event evt, Object what) {
	return false;
    }

    /**
     * By default a widget calls this function when a
     * gotFocus event occurs.
     * We strongly advise against the use of this function
     * and advise the use of doing the actions within handleEvent
     * to reduce the number of method calls.
     */
    public boolean gotFocus(Event evt, Object what) {
	return false;
    }

    /**
     * By default a widget calls this function when a
     * lostFocus event occurs.
     * We strongly advise against the use of this function
     * and advise the use of doing the actions within handleEvent
     * to reduce the number of method calls.
     */
    public boolean lostFocus(Event evt, Object what) {
	return false;
    }

    /**
     * Execute the action associated with this widget.
     */
    public void action() {
	postEvent(new Event(this, Event.ACTION_EVENT, getValue()));
    }

    /**
     * Sync the toolkit.
     */
    public void sync() {
	Toolkit.getDefaultToolkit().sync();
    }

    /**
     * Build a string describing the widget's parameters.
     * This string is used in toString.
     * @see #toString
     */
    public void paramString(StringBuffer buf) {
	if (name != null) {
	    buf.append(name);
	    buf.append(',');
	}
	buf.append(x);
	buf.append(',');
	buf.append(y);
	buf.append(',');
	buf.append(width);
	buf.append('x');
	buf.append(height);
	if ((tx != 0) || (ty != 0)) {
	    buf.append(",tx=");
	    buf.append(tx);
	    buf.append(",ty=");
	    buf.append(ty);
	}
	if (north != DETACHED) {
	    buf.append(",n=");
	    buf.append(north);
	}
	if (east != DETACHED) {
	    buf.append(",e=");
	    buf.append(east);
	}
	if (south != DETACHED) {
	    buf.append(",s=");
	    buf.append(south);
	}
	if (west != DETACHED) {
	    buf.append(",w=");
	    buf.append(west);
	}
	if (tip != null) {
	    buf.append(",tip=");
	    buf.append(tip);
	}
	if (!visible) {
	    buf.append(",hidden");
	}
	if (parent == null) {
	    buf.append(",noparent");
	}
	if (editor != null) {
	    buf.append(",editor");
	}
	if (!valid) {
	    buf.append(",invalid");
	}
	buf.append(transparent ? ",transparent" : ",opaque");
    }

    /**
     * Gets the widget class for this widget. This is useful if you
     * are interested in the actual widget class, rather than the
     * class of the script.
     */
    public Class getWidgetClass() {
	Class c = getClass();
	for (; c.getName().startsWith("script.") ; c = c.getSuperclass());
	// REMIND: remove after version 0.9
	for (; c.getName().startsWith("_") ; c = c.getSuperclass());
	return c;
    }

    /**
     * Get a short description of the widget.
     */
    public String getDescription() {
	String name = this.name;
	String clazz = getWidgetClass().getName();
	return (name == null) ? clazz : clazz + '[' + name + ']';
    }

    /**
     * Convert the widget to a string for debugging.
     * @see #paramString
     */
    public String toString() {
	StringBuffer buf = new StringBuffer();
	buf.append(getWidgetClass().getName());
	buf.append('[');
	paramString(buf);
	buf.append(']');
	return buf.toString();
    }

    /**
     * Print the widget for debugging.
     */
    public void list() {
	list(System.out, 0);
    }

    /**
     * Print the widget with indentation for debugging.
     * This function uses toString to print the widget.
     * @see #toString
     */
    public void list(PrintStream out, int n) {
	for (int i = 0 ; i < n ; i++) {
	    out.print("  ");
	}
	out.println(toString());

	for (int i = 0 ; i < nwidgets ; i++) {
	    Widget w = widgets[i];
	    if (w == null) {
		for (int j = 0 ; j < n ; j++) {
		    out.print("  ");
		}
		out.println("null");
	    } else {
		w.list(out, n+1);
	    }
	}
    }
}
