// Copyright 1996, Marimba Inc. All Rights Reserved.
// Confidential and Proprietary Information of Marimba, Inc.
// @(#)TextView.java, 1.93, 01/09/97

package marimba.text;

import java.awt.*;
import java.io.*;
import java.util.Hashtable;
import marimba.gui.*;
import marimba.util.*;
import marimba.text.keymap.*;
import marimba.persist.*;

/**
 * Abstract view of a Text object.  Multi-line and single-line
 * subclasses fill in the few methods required to make it work
 * properly in those situations.
 *
 * @author	Jonathan Payne
 * @version 1.93, 01/09/97
 */
public class TextView extends Widget implements TimerClient {
    /*
     * Additional positions to udpate when inserts and deletions
     * occur.
     */
    protected static final int TOP = 0;
    protected static final int CARET = 1;
    protected static final int SELECTION_START = 2;
    protected static final int SELECTION_END = 3;
    protected static final int NPOS = 4;

    /* Flags for text widgets. */
    protected static final int F_MODIFIED	= 1;
    protected static final int F_EDITABLE 	= 2;
    protected static final int F_BOXED    	= 4;
    protected static final int F_SELECTIONS_OK 	= 8;
    protected static final int F_SCROLLING_OK   = 16;
    protected static final int F_SAVE_TEXT	= 32;
    protected static final int F_REPAINT	= 64;
    protected static final int F_MAX_FLAG	= 64;

    /** Text we're displaying. */
    protected Text text;

    /** Insets. */
    protected Insets insets = defaultInsets;

    protected Color caretColor = Color.black;
    protected Color shadowColor = Color.white;

    /**
     * Various text position of interest which need updating as
     * characters are inserted into and deleted from the buffer.
     */
    protected int positions[] = new int[NPOS];

    /**
     * A buffer position that must be visible after the next
     * redisplay.  Ignored when it's -1.
     */
    protected int makePosVisible = -1;

    /** The current style index, while we're formatting. */
    protected int styleIndex;

    /** Flags about this widget. */
    int flags = (F_MODIFIED | F_EDITABLE | F_BOXED | F_SELECTIONS_OK
		 | F_SCROLLING_OK | F_SAVE_TEXT);

    /**
     * List of markers on this text.  Markers have positions which are
     * updated whenever inserts or deletes are made on the text.
     */
    protected Mark markList;

    /** Visible lines on the screen. */
    protected LinePool visibleLines = new LinePool();

    /** "New" lines, used for doing smart screen updates. */
    protected LinePool newLines = new LinePool();

    /** Scrollbar associated with this widget. */
    protected ScrollbarWidget sbar;

    /** Whether or not to display caret (for blinking cursor). */
    boolean caretVisible = true;

    public TextView() {
	this(new Text());
    }

    public TextView(Text text) {
	directGraphics = true;
	setText(text);
	setKeymap(defaultKeymap);
    }

    public void getProperties(PropertyList list) {
	super.getProperties(list);
	getTextProperties(list);
    }

    public void getTextProperties(PropertyList list) {
	int style = isSet(F_BOXED) ? (transparent ? BOXED : FILLED) : PLAIN;
	list.setBoolean("savetext", isSet(F_SAVE_TEXT), true);
	list.setBoolean("editable", isSet(F_EDITABLE), true);
	list.setBoolean("selections", isSet(F_SELECTIONS_OK), true);
	list.setBoolean("scrolling", isSet(F_SCROLLING_OK), true);
	list.setInteger("inset", insets.left, 2);
	list.setColor("caretcolor", caretColor, Color.black);
	list.setColor("shadowcolor", shadowColor, Color.white);
	list.setOption("style", styleOptions, style, PLAIN);
	Style def = text.styles.getDefaultStyle();
	list.setOption("align", alignOptions, def.getAlign(), LEFT);
	list.setOption("wrap", wrapOptions, def.getWrap(), Style.CHAR_WRAP);
	list.setString("text", isSet(F_SAVE_TEXT) ? text.getRichText() : null, null);
    }

    public void setProperties(PropertyList list) {
	super.setProperties(list);
	setTextProperties(list);
    }

    public void setTextProperties(PropertyList list) {
	setFlag(F_EDITABLE, list.getBoolean("editable", true));
	int style = list.getOption("style", styleOptions, PLAIN);
	boolean boxed = list.getBoolean("boxed", false);
	if (boxed) {
	    style = transparent ? BOXED: FILLED;
	} 
	setFlag(F_BOXED, style != PLAIN);
	setFlag(F_SELECTIONS_OK, list.getBoolean("selections", true));
	setFlag(F_SCROLLING_OK, list.getBoolean("scrolling", true));
	setFlag(F_SAVE_TEXT, list.getBoolean("savetext", true));
	Style def = text.styles.getDefaultStyle();
	def.setAlign(list.getOption("align", alignOptions, LEFT));
	def.setWrap(list.getOption("wrap", wrapOptions, Style.CHAR_WRAP));
	if (isSet(F_SAVE_TEXT)) {
	    text.setRichText(list.getString("text", ""));
	}
	caretColor = list.getColor("caretcolor", Color.black);
	shadowColor = list.getColor("shadowcolor", Color.white);
	int inset = list.getInteger("inset", 2);
	insets = new Insets(inset + 1, inset, inset + 1, inset);
	transparent = style != FILLED;
    }

    public static Options styleOptions = new Options();
    static {
	styleOptions.add("plain", PLAIN);
	styleOptions.add("boxed", BOXED);
	styleOptions.add("filled", FILLED);
    }

    /*
     * default insets for widget (3 on left/right to ensure room to
     * display caret when in align right mode)
     */
    protected static final Insets defaultInsets = new Insets(2, 3, 2, 3);

    /**
     * Get the value of the view as rich text
     */
    public String getRichText() {
	return text.getRichText();
    }

    /**
     * Set the value of the view as rich text
     */
    public void setRichText(String txt) {
	text.setRichText(txt);
    }

    public void setScrollbar(ScrollbarWidget sbar) {
	this.sbar = sbar;
	if (sbar != null) {
	    sbar.setProportional(false);
	    setScrollbarValues();
	}
    }

    public synchronized void invalidate() {
	setModified();
	if (visibleLines != null) {
	    visibleLines.clear();
	    newLines.clear();
	}
	super.invalidate();
    }

    /**
     * Get the current style of the text.
     * @see #setStyle
     */
    public int getStyle() {
	return isSet(F_BOXED) ? (transparent ? BOXED : FILLED) : PLAIN;
    }

    /**
     * Set the style of the text to plain, boxed or filled.
     * @see #getStyle
     */
    public void setStyle(int style) {
	switch (style) {
	  case PLAIN:
	  case BOXED:
	    setTransparent(true);
	    setBoxed(style == BOXED);
	    break;

	  case FILLED:
	    setTransparent(false);
	    setBoxed(true);
	    break;
	}
	repaint();
    }

    public Style getDefaultStyle() {
	return (Style) text.styles.getDefaultStyle().clone();
    }

    public void setDefaultStyle(Style s) {
	text.styles.setDefaultStyle(s);
    }

    public void setFontFamily(FontFamily family) {
	Style s = getDefaultStyle();
	s.setFontFamily(family);
	setDefaultStyle(s);
    }

    /**
     * Get the align of the text.
     * @see #setAlign
     */
    public int getAlign() {
	Style def = text.styles.getDefaultStyle();
	return def.getAlign();
    }

    /**
     * Set the align of the text.
     * @see #getAlign
     */
    public void setAlign(int align) {
	if (align != getAlign()) {
	    Style def = getDefaultStyle();
	    def.setAlign(align);
	    setDefaultStyle(def);
	    repaint();
	}
    }

    /**
     * Get the current wrap mode.
     * Returns a widgetConstant: CHAR, TRUNCATE or WORD.
     * @see #setWrap
     */
    public int getWrap() {
	Style  def = text.styles.getDefaultStyle();
	int  wrap = def.getWrap();
	int  w = WidgetConstants.WORD;
	switch(wrap) {
	    case Style.CHAR_WRAP:
		w = WidgetConstants.CHAR;
		break;
	    case Style.TRUNCATE: 
		w = WidgetConstants.TRUNCATE;
		break;
	    case Style.WORD_WRAP:
		w = WidgetConstants.WORD;
		break;
	}
	return w;
    }

    /**
     * Set the wrap mode to char, word, truncate.
     * A widgetConstant must be given: CHAR, TRUNCATE or WORD.
     * @see #getWrap
     */
    public void setWrap(int wrap) {
	int  w = Style.WORD_WRAP;
	switch(wrap) {
	    case WidgetConstants.CHAR:
		w = Style.CHAR_WRAP;
		break;
	    case WidgetConstants.TRUNCATE:
		w = Style.TRUNCATE;
		break;
	    case WidgetConstants.WORD:
		w = Style.WORD_WRAP;
		break;
	}
	Style def = getDefaultStyle();
	int  oldwrap = def.getWrap();
	if (oldwrap != w) {
	    def.setWrap(w);
	    setDefaultStyle(def);
	    repaint();
	}
    }

    public void validate() {
	if (!valid) {
	    Style s = getDefaultStyle();
	    s.setFont(font);
	    s.setColor(foreground);
	    setDefaultStyle(s);
	    super.validate();
	}
    }

    public String getEditor() {
	return "marimba.builder.RichTextBoxWidgetEditor";
    }

    /** Set this view's text object. */
    public void setText(Text text) {
	if (this.text == text)
	    return;
	if (this.text != null)
	    this.text.removeView(this);
	this.text = text;
	if (text != null) {
	    text.addView(this);
	}
	setModified();
    }

    final protected void addFlag(int bits) {
	flags |= bits;
    }

    final protected void setFlag(int bits, boolean on) {
	if (on)
	    addFlag(bits);
	else
	    clearFlag(bits);
    }

    final protected void clearFlag(int bits) {
	flags &= ~bits;
    }

    final protected boolean isSet(int bits) {
	return ((flags & bits) != 0);
    }

    /**
     * Updates the line structures and styles list based on LENGTH
     * characters being inserted at POS.
     */
    protected void notifyInsert(int pos, int length) {
	for (int i = 0; i < NPOS; i++) {
	    if (pos < positions[i]) {
		positions[i] += length;
	    }
	}
	for (Mark m = markList; m != null; m = m.next) {
	    if (pos < m.pos) {
		m.pos += length;
	    }
	}
	visibleLines.notifyInsert(pos, length);
	newLines.notifyInsert(pos, length);
	setModified();
	initCaret();
    }

    /**
     * Updates the line structures and styles list based on LENGTH
     * characters being deleted from position POS.
     */
    protected void notifyDelete(int pos, int length) {
	int endpos = pos + length;

	for (int i = 0; i < NPOS; i++) {
	    if (endpos <= positions[i]) {
		positions[i] -= length;
	    } else if (pos < positions[i] && endpos >= positions[i]) {
		positions[i] = pos;
	    }
	}
	for (Mark m = markList; m != null; m = m.next) {
	    if (endpos <= m.pos) {
		m.pos -= length;
	    } else if (pos < m.pos && endpos >= m.pos) {
		m.pos = pos;
	    }
	}
	visibleLines.notifyDelete(pos, length);
	newLines.notifyDelete(pos, length);
	setModified();
	initCaret();
    }

    /**
     * Text object notifies us that a region of text has changed,
     * probably because of a style change.
     */
    protected void notifyDirtyRegion(int pos0, int pos1) {
	dirtyRegion(pos0, pos1, false);
    }

    protected void dirtyRegion(int pos0, int pos1, boolean notIfEmpty) {
	if (pos0 > pos1) {
	    int tmp = pos1;
	    pos1 = pos0;
	    pos0 = tmp;
	} else if (notIfEmpty && pos0 == pos1)
	    return;
	if (visibleLines != null) {
	    visibleLines.dirtyRegion(pos0, pos1);
	    newLines.dirtyRegion(pos0, pos1);
	    setModified();
	}
    }

    protected void repaintRegion(int pos0, int pos1, boolean notIfEmpty) {
	if (pos0 > pos1) {
	    int tmp = pos1;
	    pos1 = pos0;
	    pos0 = tmp;
	} else if (notIfEmpty && pos0 == pos1) {
	    return;
	}
	if (visibleLines != null) {
	    visibleLines.dirtyRegion(pos0, pos1);
	}
	repaint();
    }

    /** Locates a line object based on a Y position. */
    protected Line locateLineContainingY(int y) {
	return visibleLines.lineContainingY(y);
    }


/******************** Layout **********************/

    /**
     * Returns the buffer position that corresponds to the specified
     * x/y position.
     */
    public int locateHit(int x, int y) {
	Line line = locateLineContainingY(y);

	if (line == null)
	    return text.length();
	if (y > line.y + line.height)
	    return Math.min(line.pos1, text.count);
	return scanLine(line, null, LOCATE, (float) x, line.posV);
    }

    protected final int measureLine(Line line, float xlimit, int poslimit) {
	float spaceShim;
	int sindex;
	float x;
	Style style;
	StylePool styles = text.styles;

	line.spaceShim = spaceShim = 0;
	sindex = line.styleIndex = styleIndex;
	style = styles.getStyleAt(sindex);
	x = line.x = insets.left + style.leftMargin;

	byte data[] = text.data;
	int pos = line.pos;
	float newx;
	int nextStylePos;
	boolean hitNewline = false;
	int nspace;
	int ntabs;
	float x0;
	int pos0;
	int lineHeight;
	int lineAscent;
	int breakLineHeight;
	int breakLineAscent;
	int breakPos;
	float breakX;

	DeviceFont font = style.getFont();
	int widths[] = font.getWidths();
	FontMetrics metrics = font.getMetrics();
	lineHeight = metrics.getHeight();
	lineAscent = metrics.getAscent();

	nspace = ntabs = breakLineHeight = breakLineAscent = breakPos = 0;
	newx = breakX = (float) 0.;

	while (pos < poslimit) {
	    int spaceCount = 0;

	    nextStylePos = styles.getStylePosAt(sindex + 1);
	    if (nextStylePos > poslimit)
		nextStylePos = poslimit;

	    int tabWidth = widths[' '] * 8;

	    x0 = x;
	    pos0 = pos;

	    /* scan until we run out of X or run out of buffer
	       or we run into the next style change */
	  loop:
	    while (pos < nextStylePos) {
		int c = data[pos];

		switch (c) {
		  case '\r':
		  case '\n':
		    hitNewline = true;
		    break loop;

		  case '\t':
		    newx = x + tabWidth
			- ((x - style.leftMargin - insets.left) % tabWidth);
		    if (newx > xlimit)
			break loop;
		    int h = metrics.getHeight();
		    if ((style.fontFace & FontFamily.EMBOSS) != 0)
			h += 1;
		    lineHeight = Math.max(h, lineHeight);
		    lineAscent = Math.max(metrics.getAscent(), lineAscent);
		    x0 = newx;
		    pos0 = pos + 1;
		    spaceCount = nspace = 0;
		    break;

		  case ' ':
		    spaceCount += 1;
		    breakPos = pos;
		    breakX = x;
		    newx = x + widths[' '];
		    if (newx > xlimit)
			break loop;
		    break;

		  default:
		    newx = x + widths[c & 0xFF];
		    if (newx > xlimit)
			break loop;
		}
		x = newx;
		pos += 1;
	    }
	    int diff = pos - pos0;
	    if (diff > 0 || hitNewline) {
		int h = metrics.getHeight();
		int a = metrics.getAscent();
		if ((style.fontFace & FontFamily.EMBOSS) != 0)
		    h += 1;
		if (pos0 == line.pos) {
		    /* if this substring started at the beginning of the
		       line, then this is the initial height of the line */
		    lineHeight = h;
		    lineAscent = a;
		} else {
		    lineHeight = Math.max(h, lineHeight);
		    lineAscent = Math.max(a, lineAscent);
		}
	    }

	    /* If there were any spaces during this font change, then
	       record the lineAscent and lineHeight which *must* have
	       been in effect at that space in case we have to back up
	       for word wrap mode. */
	    if (spaceCount > 0) {
		nspace += spaceCount;
		breakLineAscent = lineAscent;
		breakLineHeight = lineHeight;
	    }
	    if (hitNewline || newx > xlimit || pos == poslimit)
		break;
	    sindex += 1;
	    style = styles.getStyleAt(sindex);
	    font = style.getFont();
	    widths = font.getWidths();
	    metrics = font.getMetrics();
	}

	int eolPos = pos;
	if (hitNewline) {
	    if (pos < poslimit + 1 && data[pos] == '\r'
		&& data[pos + 1] == '\n')
		pos += 2;
	    else
		pos += 1;
	}
	switch (style.wrapStyle) {
	  default:
	  case Style.CHAR_WRAP:
	    line.posE = line.pos1 = pos;
	    line.posV = eolPos;
	    break;

	  case Style.TRUNCATE:
	    line.posV = eolPos;
	    line.posV = pos;
	    if (!hitNewline) {
		while (pos < poslimit && data[pos] != '\n'
		       && data[pos] != '\r')
		    pos += 1;
		if (pos < poslimit && data[pos] == '\r')
		    pos += 1;
		if (pos < poslimit && data[pos] == '\n')
		    pos += 1;
	    }
	    line.posE = line.pos1 = pos;
	    break;

	  case Style.WORD_WRAP:
	    if (nspace > 0 && newx > xlimit) {
		line.posE = pos;
		pos = line.posV = breakPos;
		x = breakX;
		lineHeight = breakLineHeight;
		lineAscent = breakLineAscent;
		while (styles.getStylePosAt(sindex) > pos)
		    sindex -= 1;
		while (pos < poslimit && data[pos] == ' ')
		    pos += 1;
		while (pos >= styles.getStylePosAt(sindex + 1))
		    sindex += 1;
		line.pos1 = pos;
	    } else {
		line.posE = line.pos1 = pos;
		line.posV = eolPos;
	    }
	    break;
	}
	switch (style.justifyStyle) {
	  case LEFT:
	    break;

	  case RIGHT:
	    line.x += (xlimit - x);
	    break;

	  case CENTER:
	    line.x += (xlimit - x) / 2;
	    break;

	  case FILL:
	    if (nspace > 1 && newx > xlimit)
		line.spaceShim = (float) (xlimit - x) / (nspace - 1);
	    break;
	}
	line.height = (short) lineHeight;
	line.ascent = (short) lineAscent;
	styleIndex = sindex;
	return pos;
    }

    protected final int OUTPUT = 1;
    protected final int LOCATE = 2;
    protected final int POSITION = 3;

    private boolean paintingInSelection; 

    /**
     * This is a duplicate of the measureLine method without the
     * measuring overhead.  This was separated out to make painting,
     * position and x/y location faster, since they are very common
     * operations.
     */
    protected final int scanLine(Line line, Graphics g, int what,
				 float xlimit, int poslimit) {
	float spaceShim;
	int sindex;
	float x;
	Style style;
	StylePool styles = text.styles;

	if (line.tabCount > 0) {
	    spaceShim = 0;
	} else {
	    spaceShim = line.spaceShim;
	}
	sindex = line.styleIndex;
	style = styles.getStyleAt(sindex);
	x = line.x;
	if (poslimit > line.posV) {
	    poslimit = line.posV;
	}

	byte data[] = text.data;
	int pos = line.pos;
	int ntabs;
	float x0 = (float) 0;
	float newx = (float) 0;
	int pos0;
	boolean hitNewline = false;

	ntabs = 0;

      loop:
	while (pos < poslimit) {
	    DeviceFont font = style.getFont();
	    int widths[] = font.getWidths();
	    FontMetrics metrics = font.getMetrics();
	    int nextStylePos;

	    nextStylePos = styles.getStylePosAt(sindex + 1);
	    if (nextStylePos > poslimit)
		nextStylePos = poslimit;

	    if (what == OUTPUT) {
		g.setColor(paintingInSelection ? getSelForeground() :
			   style.color);
		g.setFont(font);
	    }

	    int tabWidth = widths[' '] * 8;

	    x0 = x;
	    pos0 = pos;

	    while (pos < nextStylePos) {
		int c = data[pos];

		switch (c) {
		  case '\r':
		  case '\n':
		    hitNewline = true;
		    break;

		  case '\t':
		    newx = x + tabWidth
			- ((x - style.leftMargin - insets.left) % tabWidth);
		    if (newx > xlimit)
			break;
		    if (++ntabs == line.tabCount)
			spaceShim = line.spaceShim;
		    break;

		  case ' ':
		    newx = x + widths[' '];
		    if (spaceShim > 0) {
			newx += spaceShim;
			break;
		    }
		    if (newx > xlimit)
			break;
		    x = newx;
		    pos += 1;
		    continue;

		  default:
		    newx = x + widths[c & 0xFF];
		    if (newx > xlimit)
			break;
		    x = newx;
		    pos += 1;
		    continue;
		}
		if (what == OUTPUT && pos > pos0) {
		    if ((style.fontFace & FontFamily.EMBOSS) != 0) {
			g.setColor(shadowColor);
			g.drawBytes(data, pos0, pos - pos0,
				    (int) x0 + 1, line.y + line.ascent + 1);
			g.setColor(style.color);
		    }
		    g.drawBytes(data, pos0, pos - pos0, (int) x0, line.y + line.ascent);
		}
		if (hitNewline || newx > xlimit)
		    break loop;
		x0 = x = newx;
		pos0 = ++pos;
	    }
	    if (what == OUTPUT && pos > pos0) {
		if ((style.fontFace & FontFamily.EMBOSS) != 0) {
		    g.setColor(shadowColor);
		    g.drawBytes(data, pos0, pos - pos0,
				(int) x0 + 1, line.y + line.ascent + 1);
		    g.setColor(style.color);
		}
		g.drawBytes(data, pos0, pos - pos0, (int) x0, line.y + line.ascent);
	    }
	    if (pos == poslimit)
		break;
	    sindex += 1;
	    style = styles.getStyleAt(sindex);
	}

	/* Now figure out what to return. */
	switch (what) {
	  case LOCATE:
	    if (pos < line.posV && !hitNewline && xlimit > (x + newx) / 2)
		pos += 1;
	    return pos;

	  case POSITION:
	    return (int) x;

	  default:
	  case OUTPUT:
	    return 0;
	}
    }

    /**
     * Paint a line.  The work is done by scanLine, but this does a
     * little work itself to figure out selection stuff.  When a line
     * is partially selected, the backgrounds are drawn first, and
     * then the text is drawn on top.  Portions of this are kludgy
     * because I didn't completely separate painting a selection and
     * painting a caret.  They are slightly different cases.
     */
    protected void paintLine(Graphics g, Line line, boolean doSelections) {
	int s0 = positions[SELECTION_START];
	int s1 = positions[SELECTION_END];
	int lineStart = line.pos;
	int lineEnd = line.pos1;
	boolean isCaret = (s0 == s1);

	if (!doSelections || s0 >= lineEnd || s1 < lineStart
	    || (isCaret && s0 == lineEnd  // we'll catch it on next line ...
		&& s0 != lineStart	  // ... unless this is the last line
		                          // that's empty
		   )) {
	    /* completely not selected, or we're not showing selections */
	    if (!transparent) {
		g.setColor(hilite);
		g.fillRect(0, line.y, width, line.height);
	    }
	    paintingInSelection = false;
	    scanLine(line, g, OUTPUT, Float.MAX_VALUE, line.posV);
	} else if (s0 <= lineStart && s1 > lineEnd) {
	    /* completely selected */
	    g.setColor(getSelBackground());
	    g.fillRect(insets.left, line.y, width - insets.left - insets.right,
		       line.height);
	    paintingInSelection = true;
	    scanLine(line, g, OUTPUT, Float.MAX_VALUE, line.posV);
	} else {
	    /* line is partially selected */
	    int x0 = -1;
	    int x1 = -1;
	    Color sc = isCaret ? caretColor : getSelBackground();

	    if (s0 >= lineStart && s0 <= lineEnd) {
		x0 = scanLine(line, g, POSITION, Float.MAX_VALUE, s0);
	    }
	    if (isCaret) {
		x1 = x0 + 2;
	    } else if (s1 >= lineStart && s1 < lineEnd) {
		/* selection is a real selection */
		x1 = scanLine(line, g, POSITION, Float.MAX_VALUE, s1);
	    }
	    if (!transparent) {
		/* first clear the non-selection parts */
		g.setColor(hilite);
		g.fillRect(insets.left, line.y, width - insets.left,
			   line.height);
	    }

	    /* paint entire line unselected */
	    paintingInSelection = false;
	    scanLine(line, g, OUTPUT, Float.MAX_VALUE, line.posV);

	    /* now draw the selection part */
	    if (x0 < 0) {
		x0 = insets.left;
	    }
	    if (x1 < 0) {
		x1 = width - insets.right;
	    }
	    /* draw the selection if it's not a caret or (if it is a caret)
	       the widget is currently editable for one reason or another */
	    if ((!isCaret || editor != null || isSet(F_EDITABLE)) && x0 < x1) {
		g.setColor(sc);
		g.fillRect(x0, line.y, x1 - x0, line.height);
	    }
	    if (!isCaret) {
		paintingInSelection = true;
		Graphics gc = g.create();
		try {
		    gc.clipRect(x0, line.y, x1 - x0, line.height);
		    scanLine(line, gc, OUTPUT, Float.MAX_VALUE, line.posV);
		} finally {
		    gc.dispose();
		}
	    }
	}
    }

    /**
     * Make newLines reflect what the view should look like after the
     * next redisplay.  This may borrow lines from visible lines to
     * speed up the updating.
     */
    boolean layoutLines() {
	int pos = positions[TOP];
	int y = insets.top;
	int poslimit = text.length();
	int ylimit = height - insets.bottom;
	boolean tryrop = false;
	LinePool visibleLines = this.visibleLines;
	LinePool newLines = this.newLines;
	StylePool styles = text.styles;
	int lineIndex = 0;
	Style style;

	styleIndex = styles.findStyleIndexFor(pos);
	while (pos <= poslimit) {
	    Line nline = newLines.getLineAt(lineIndex);
	    Line vline = visibleLines.getLineAt(lineIndex++);
	    int last = pos;

	    style = styles.getStyleAt(styleIndex);
	    float xlimit = width - style.rightMargin - insets.right;
	    if (nline.pos != pos || nline.modified) {
		Line l;

		/* first see if we can find the line from the previous
		   redisplay, but only if the reason we're reformatting
		   is that the pos is different */
		if (!nline.modified
		    && (l = visibleLines.findLineAt(pos)) != null) {
		    nline.copy(l);
		    pos = nline.pos1;
		    tryrop = true;
		    nline.ymodified = true;
		    nline.styleIndex = styleIndex;
		} else {
		    nline.pos = pos;
		    pos = measureLine(nline, xlimit, poslimit);
		}
		nline.modified = true;
		nline.y = y;
	    } else {
		pos = nline.pos1;
		if (nline.y != y) {
		    nline.modified = nline.ymodified = true;
		    nline.y = y;
		    tryrop = true;
		}
		/* Changes above might have changed the style index.
		   Make note of that now. */
		nline.styleIndex = styleIndex;
	    }
	    while (pos >= styles.getStylePosAt(styleIndex + 1))
		styleIndex += 1;
	    y += nline.height;
	    if (pos == poslimit && (pos == 0 || text.data[pos - 1] != '\n')) {
		nline.pos1 += 1;
		break;
	    }
	    if (last == pos) {
		nline.pos1 += 1;
		break;
	    }
	    if (y >= ylimit)
		break;
	}
	newLines.finishLines(lineIndex);

	return tryrop;
    }
	
    /**
     * Try to do smart updating of screen by just moving around lines,
     * using copyarea.  Lines that are candidates are ones that are
     * not modified, but whose Y positions are different.
     */
    void copyLines(Graphics g, int vfrom, int nfrom) {
	Line vlines[] = visibleLines.lines;
	Line nlines[] = newLines.lines;
	Line v, n;
	int vlimit = visibleLines.lineCount;
	int nlimit = newLines.lineCount;
	int ylimit = height - insets.bottom;

	while (vfrom < vlimit && nfrom < nlimit) {
	    /* skip lines which are already in the right place */
	    while (vfrom < vlimit && nfrom < nlimit) {
		v = vlines[vfrom];
		n = nlines[nfrom];

		if (n.modified || n.ymodified)
		    break;
		vfrom += 1;
		nfrom += 1;
	    }

	    /* Now we're at the first line which could be moved.  Let's
	       get the line.pos values lined up. */
	    while (vfrom < vlimit && nfrom < nlimit) {
		v = vlines[vfrom];
		n = nlines[nfrom];

		if (v.pos < n.pos)
		    vfrom += 1;
		else
		    break;
	    }

	    while (vfrom < vlimit && nfrom < nlimit) {
		v = vlines[vfrom];
		n = nlines[nfrom];

		if (n.pos < v.pos)
		    nfrom += 1;
		else
		    break;
	    }

	    int count = 0;
	    /* Now find length of region of lines to be moved.  Only
	       lines that are ymodified can be considered for moving. */
	    while (vfrom < vlimit && nfrom < nlimit) {
		v = vlines[vfrom];
		n = nlines[nfrom];
		if (v.pos != n.pos || !n.ymodified
		    || v.y + v.height >= ylimit
		    || n.y + n.height >= ylimit) {
		    break;
		}
		count += 1;
		vfrom += 1;
		nfrom += 1;
	    }
	    if (count > 0) {
		int ydiff = vlines[vfrom - 1].y - nlines[nfrom - 1].y;

		Line first = vlines[vfrom - count];
		Line last = vlines[vfrom - 1];
		if (ydiff < 0) {
		    copyLines(g, vfrom, nfrom);
		    g.copyArea(0, first.y, width,
			       last.y + last.height - first.y,
			       0, -ydiff);
		} else if (ydiff > 0) {
		    g.copyArea(0, first.y, width,
			       last.y + last.height - first.y,
			       0, -ydiff);
		    copyLines(g, vfrom, nfrom);
		}
		while (count > 0) {
		    Line nline = nlines[nfrom - count];
		    nline.modified = nline.ymodified = false;
		    vlines[nfrom - count].copy(nline);
		    count -= 1;
		}
	    }
	    vfrom += 1;
	    nfrom += 1;
	}
    }
	
    /** Y location of the bottom-most line after last redisplay. */
    int bottomY;

    /** Paint all the layed out lines now. */
    void paintLines(Graphics g, boolean damage) {
	int limit = visibleLines.lineCount;

	g.clipRect(insets.left, insets.top,
		   width - insets.left - insets.right,
		   height - insets.top - insets.bottom);

	Line line = null;
	boolean doSelections;
	doSelections = (((editor != null) ||
			 isSet(F_SELECTIONS_OK | F_EDITABLE)) &&
			(caretVisible || selectionLength() > 0));
	for (int i = 0; i < limit; i++) {
	    line = visibleLines.getLineAt(i);

	    if (line.modified || damage) {
		//System.out.println("Painting " + i);
		paintLine(g, line, doSelections);
		line.ymodified = line.modified = false;
		newLines.getLineAt(i).copy(line);
	    }
	}
	int lastY = bottomY;
	if (line != null) {
	    bottomY = line.y + line.height;
	}
	if (!transparent && lastY > bottomY) {
	    g.setColor(hilite);
	    g.fillRect(0, bottomY, width, lastY - bottomY);
	}
	newLines.finishLines(limit);
	if (sbar != null) {
	    //REMIND: This is called FAR too often!
	    //
	    setScrollbarValues();
	}
	clearFlag(F_REPAINT);
    }

    protected void setScrollbarValues() {
	if (visibleLines != null && text != null) {
	    sbar.setParam(visibleLines.pos0, 0, text.length() - 1,
			  0,
			  //visibleLines.getPageSize(this)
			  1);
	}
    }

    protected boolean posVisible(int pos) {
	return visibleLines.posVisible(pos, this);
    }

    protected void setModified() {
	if (!isSet(F_MODIFIED)) {
	    synchronized (this) {
		addFlag(F_MODIFIED);
		if (!isSet(F_REPAINT)) {
		    addFlag(F_REPAINT);
		    super.repaint();
		}
	    }
	}
    }

    public synchronized void repaint() {
	if (!isSet(F_REPAINT)) {
	    addFlag(F_REPAINT);
	    super.repaint();
	}
    }

/****************** Selections *********************/

    /**
     * Sets the selection, expanding it to words or lines if level is
     * 2 or 3 respectively.  pos1 is where the mouse is, which is where
     * the caret should end up.
     */
    public void select(int pos0, int pos1, int level) {
	boolean switched = (pos0 > pos1);
	if (switched) {
	    int tmp = pos0;
	    pos0 = pos1;
	    pos1 = tmp;
	}
	switch (level) {
	  case 0:
	  case 1:
	    break;

	  case 2:
	    pos0 = text.getWordEnd(pos0, -1);
	    pos1 = text.getWordEnd(pos1, 1);
	    break;

	  case 3:
	    pos0 = text.getBeginningOfLine(pos0);
	    pos1 = text.getEndOfLine(pos1);
	    if (pos1 < text.length())
		pos1 += 1;		/* select the whole line */
	    break;

	  case 4:
	  default:
	    pos0 = 0;
	    pos1 = text.length();
	    break;
	}
	select(pos0, pos1);
	positions[CARET] = switched ? pos0 : pos1;
    }

    public synchronized void select(int pos0, int pos1) {
	int p[] = positions;

	if (pos0 > pos1) {
	    int tmp = pos0;
	    pos0 = pos1;
	    pos1 = tmp;
	}
	if (pos0 != p[SELECTION_START] || pos1 != p[SELECTION_END]) {
	    if (p[SELECTION_END] < pos0
		|| p[SELECTION_START] > pos1) {
		/* no intersection */
		repaintRegion(p[SELECTION_START], p[SELECTION_END], false);
		repaintRegion(pos0, pos1, false);
	    } else {
		repaintRegion(p[SELECTION_START], pos0, true);
		repaintRegion(p[SELECTION_END], pos1, true);
	    }
	    p[SELECTION_START] = pos0;
	    p[SELECTION_END] = pos1;
	}
    }

    public synchronized void extendSelection(int pos) {
	select(getSelectionPivot(), pos);
	positions[CARET] = pos;
	makePosVisible = pos;
    }

    public synchronized int getSelectionPivot() {
	return (positions[CARET] == positions[SELECTION_END]) ?
	    positions[SELECTION_START] : positions[SELECTION_END];
    }

    public int selectionLength() {
	return positions[SELECTION_END] - positions[SELECTION_START];
    }

    public String getSelection() {
	return text.substring(positions[SELECTION_START],
			      positions[SELECTION_END]);
    }

    public final int getSelectionStart() {
	return positions[SELECTION_START];
    }

    public final int getSelectionEnd() {
	return positions[SELECTION_END];
    }

    public void setCaret(int caret) {
	if (caret < 0) {
	    caret = 0;
	} else if (caret > text.length()) {
	    caret = text.length();
	}
	select(caret, caret);
	positions[CARET] = caret;
	makePosVisible = caret;
	initCaret();
    }

    final public int getCaret() {
	return positions[CARET];
    }

    /**
     * Set the top position of the view.  This is how scrolling,
     * whether horizontal or vertical, actually occurs.
     */
    void setTop(int top) {
	if (top != positions[TOP]) {
	    positions[TOP] = top;
	    setModified();
	}
    }

    public Mark getMark(int pos) {
	Mark m = new Mark(this, pos);
	m.next = markList;
	markList = m;

	return m;
    }

    void clearMarks() {
	Mark m = markList;
	markList = null;
	while (m != null) {
	    m.view = null;
	    m = m.next;
	}
    }


/****************** Get and set methods *******************/

    /**
     * Get the color of the caret.
     */
    public Color getCaretColor() {
	return caretColor;
    }

    /**
     * Set the color of the caret.
     */
    public void setCaretColor(Color caretColor) {
	if (caretColor == null || !caretColor.equals(this.caretColor)) {
	    this.caretColor = caretColor;
	    repaint();
	}
    }

    /**
     * Get the shadow color.
     */
    public Color getShadowColor() {
	return shadowColor;
    }

    /**
     * Set the shadow color.
     */
    public void setShadowColor(Color shadowColor) {
	if (shadowColor == null || !shadowColor.equals(this.shadowColor)) {
	    this.shadowColor = shadowColor;
	    repaint();
	}
    }

    public void setEditable(boolean on) {
	setFlag(F_EDITABLE, on);
    }

    public void setBoxed(boolean on) {
	setFlag(F_BOXED, on);
    }

    public void setSelections(boolean on) {
	setFlag(F_SELECTIONS_OK, on);
    }

    public void setScrolling(boolean on) {
	setFlag(F_SCROLLING_OK, on);
    }

    public void setSavetext(boolean on) {
	setFlag(F_SAVE_TEXT, on);
    }

    public void setInsets(Insets insets) {
	this.insets = insets;
	repaint();
    }	

    public void setInset(int inset) {
	if (inset >= 0 && inset != getInset()) {
	    insets = new Insets(inset + 1, inset, inset + 1, inset);
	    repaint();
	}
    }

    public boolean getEditable() {
	return isSet(F_EDITABLE);
    }

    public boolean getBoxed() {
	return isSet(F_BOXED);
    }

    public boolean getSelections() {
	return isSet(F_SELECTIONS_OK);
    }

    public boolean getScrolling() {
	return isSet(F_SCROLLING_OK);
    }

    public boolean getSavetext() {
	return isSet(F_SAVE_TEXT);
    }

    public Insets getInsets() {
	return insets;
    }

    public int getInset() {
	return insets.left;
    }

    /** Get this view's text object. */
    public Text getTextValue() {
	return text;
    }

    /** Return the text as a string. */
    public String getStringValue() {
	return getText();
    }

    /** Return the text as a string. */
    public Object getValue() {
	return getText();
    }

    /** Set the text to the specified string. */
    public synchronized void setValue(Object value) {
	if (value instanceof String) {
	    setText((String)value);
	} else if (value instanceof Text) {
	    setText((Text)value);
	} else if (value == null) {
	    setText("");
	} else {
	    setText(String.valueOf(value));
	}
    }

    /** Return the text as a string. */
    public String getText() {
	return text.toString();
    }

    /** Set the text to the specified string. */
    public synchronized void setText(String value) {
	text.clear();
	clearMarks();
	if (value != null) {
	    text.insert(value, 0, value.length());
	}
    }

/****************** Keyboard input handling. **********************/

    public boolean focusInterest() {
	return isSet(F_EDITABLE);
    }

    public void deleteSelection() {
	if (selectionLength() > 0) {
	    delete(1);
	}
    }

    public void insert(int key) {
	if ((editor == null) && !isSet(F_EDITABLE)) {
	    return;
	}
	deleteSelection();
	text.insert(key, getCaret());
	setCaret(getCaret() + 1);
    }

    public void insert(String s) {
	if ((editor == null) && !isSet(F_EDITABLE)) {
	    return;
	}
	deleteSelection();
	text.insert(s, getCaret(), s.length());
	setCaret(getCaret() + s.length());
    }

    /** Delete from the caret to the specified position. */
    public void deleteTo(int pos) {
	delete(pos - getCaret());
    }
	
    public void delete(int arg) {
	if ((editor == null) && !isSet(F_EDITABLE)) {
	    return;
	}
	int caret = getCaret();

	if (selectionLength() > 0)
	    text.delete(getSelectionStart(), selectionLength());
	else {
	    if (caret + arg < 0 || caret + arg > text.length())
		return;
	    text.delete(caret, arg);
	    if (arg < 0) {
		setCaret(caret + arg);
		makePosVisible = caret + arg - 1;
	    }
	}
    }

    /** The root keymap for this view. */
    Keymap keymap = defaultKeymap;

    /** The current piece of the keymap tree we're looking at. */
    Keymap keymapState = null;

    public void setKeymap(Keymap map) {
	keymap = keymapState = map;
    }

    public Keymap getSettableKeymap() {
	return keymap = Keymap.getSparseKeymap(defaultKeymap);
    }

    boolean processKey(Event e, int key) {
	if (keymapState == null) {
	    keymapState = keymap;
	}
	KeymapItem next = keymapState.getKeyParent(key);
	if (next == null) {
	    keymapState = null;
	    if (e.id == Event.KEY_PRESS) {
		insert(e.key);
	    }
	} else if (next.isKeymap()) {
	    keymapState = (Keymap) next;
	} else {
	    keymapState = null;
	    if (next instanceof KeyboardCommand) {
		((KeyboardCommand) next).execute(this, e);
	    }
	}
	return (next != null);	    
    }

    int lineColumn = -1;

    public boolean handleEvent(Event evt) {
	int pos = -1;
	int caret = getCaret();
	switch (evt.id) {
	  case Event.MOUSE_ENTER:
	  case Event.MOUSE_EXIT:
	  case Event.MOUSE_MOVE:
	    break;

	  case Event.KEY_PRESS:
	  case Event.KEY_ACTION:
	    /* handle key combinations that change the focus */
	    if (evt.key == '\t') {
		if (evt.controlDown() && evt.metaDown()) {
		    if (evt.shiftDown()) {
			previousFocus();
		    } else {
			nextFocus();
		    }
		    return true;
		} else if (evt.controlDown()) {
		    /* focus on next/previous page in folder */
		    /* let default focus behavior take care of this */
		    return false;
		}
	    }

	    /* handle keyboard combinations that extend the selection */
	    if (evt.modifiers == Event.SHIFT_MASK) {
		switch (evt.key) {
		  case Event.LEFT:
		    pos = Math.max(caret - 1, 0); break;
		  case Event.RIGHT:
		    pos = Math.min(caret + 1, text.length()); break;
		  case Event.END: {
		    Line l = visibleLines.lineContainingPos(caret);
		    pos = (l == null) ? text.length() : l.posV;
		    break;
		  }
		  case Event.HOME: {
		    Line l = visibleLines.lineContainingPos(caret);
		    pos = (l == null) ? 0 : l.pos;
		    break;
		  }
		  case Event.UP:
		  case Event.DOWN: {
		      if (lineColumn == -1) {
			  lineColumn = text.getColumnFor(caret);
		      }
		      int src = (evt.key == Event.UP) ?
			  text.getBeginningOfLine(caret) - 1 :
			  text.getNextLine(caret);
		      if (src >= 0) {
			  pos = text.getPosFor(lineColumn, src);
		      } else {
			  pos = 0;
		      }
		      break;
		  }
		}
		if (pos != -1) {
		    extendSelection(pos);
		    return true;
		}
	    }
	    if (evt.modifiers == (Event.SHIFT_MASK | Event.CTRL_MASK)) {
		switch (evt.key) {
		  case Event.HOME:
		    pos = 0; break;
		  case Event.END:
		    pos = text.length(); break;
		  case Event.LEFT:
		    pos = text.getNextWord(caret, -1); break;
		  case Event.RIGHT:
		    pos = text.getNextWordStart(caret); break;
		}
		if (pos != -1) {
		    extendSelection(pos);
		    return true;
		}
	    }

	    /* not a selection command, so do normal keymap stuff */
	    if (evt.metaDown()) {
		processKey(evt, Keymap.META);
	    }
	    if (evt.controlDown()) {
		processKey(evt, Keymap.CONTROL);
	    }
	    return processKey(evt, evt.key);

	  case Event.GOT_FOCUS:
	    initCaret();
	    return false;

	  case Event.LOST_FOCUS:
	    Timer.master.remove(this);
	    if (caretVisible) {
		tick(0, null);
	    }
	    return false;

	  default:
	    if (!(evt.arg instanceof KeyboardCommand)) {
		return super.handleEvent(evt);
	    }
	    handleCommand(evt, (KeyboardCommand) evt.arg);
	    return true;
	}
	return super.handleEvent(evt);
    }

    protected void handleCommand(Event evt, KeyboardCommand cmd) {
	boolean lineCmd = false;

	switch (cmd.getCode()) {
	  case CUT:
	    if (selectionLength() > 0) {
		Environment.setClipboard(getSelection());
	    }
	    deleteSelection();
	    break;

	  case COPY:
	    if (selectionLength() > 0) {
		Environment.setClipboard(getSelection());
	    }
	    break;

	  case PASTE: {
	      Object clip;

	      clip = Environment.getClipboard();
	      if (clip != null && clip instanceof String) {
		  insert((String)clip);
	      }
	      break;
	  }

	  case NEWLINE_AND_BACKUP:
	    insert('\n');
	    setCaret(getCaret() - 1);
	    break;

	  case SET_MARK:
	    break;

	  case BACKWARD_CHAR:
	    setCaret(getCaret() - 1);
	    break;

	  case BACKWARD_WORD:
	    setCaret(text.getNextWord(getCaret(), -1));
	    break;

	  case BEGINNING_OF_LINE:
	    setCaret(text.getBeginningOfLine(getCaret()));
	    break;

	  case END_OF_LINE:
	    setCaret(text.getEndOfLine(getCaret()));
	    break;

	  case BEGINNING_OF_VIEW_LINE:
	  case END_OF_VIEW_LINE: {
	      Line l = visibleLines.lineContainingPos(getCaret());
	      if (l != null) {
		  setCaret((cmd.getCode() == BEGINNING_OF_VIEW_LINE) ?
			   l.pos : l.posV);
	      }
	      break;
	  }

	  case FORWARD_CHAR:
	    setCaret(getCaret() + 1);
	    break;

	  case FORWARD_WORD:
	    /* foward word EMACS style */
	    setCaret(text.getNextWord(getCaret(), 1));
	    break;

	  case FORWARD_WORD_START:
	    /* foward word PC style */
	    setCaret(text.getNextWordStart(getCaret()));
	    break;

	  case KILL_TO_EOL: {
	      int caret = getCaret();
	      int pos = text.getEndOfLine(caret);
	      if (pos == caret)
		  pos += 1;
	      deleteTo(pos);
	      break;
	  }

	  case NEXT_PAGE:
	    scrollPage(1, true);
	    break;

	  case PREVIOUS_PAGE:
	    scrollPage(-1, true);
	    break;

	  case SELF_INSERT:
	    insert(evt.key);
	    break;

	  case DELETE_PREVIOUS_CHAR:
	    delete(-1);
	    break;

	  case DELETE_NEXT_CHAR:
	    delete(1);
	    break;

	  case BEGINNING_OF_BUFFER:
	    setCaret(0);
	    break;

	  case END_OF_BUFFER:
	    setCaret(text.length());
	    break;

	  case DELETE_NEXT_WORD:
	    deleteTo(text.getNextWord(getCaret(), 1));
	    break;

	  case DELETE_PREVIOUS_WORD:
	    deleteTo(text.getNextWord(getCaret(), -1));
	    break;

	  case MAKE_PLAIN:
	    applyStyleChange(new StyleChange().makePlain());
	    break;

	  case MAKE_ITALICS:
	    applyStyleChange(new StyleChange().makeItalic());
	    break;

	  case MAKE_BOLD:
	    applyStyleChange(new StyleChange().makeBold());
	    break;

	  case MAKE_EMBOSS:
	    applyStyleChange(new StyleChange().makeEmbossed());
	    break;

	  case MAKE_LARGER:
	    applyStyleChange(new StyleChange().changeFontSize(1));
	    break;

	  case MAKE_SMALLER:
	    applyStyleChange(new StyleChange().changeFontSize(-1));
	    break;

	  case NEXT_LINE:
	    handleNextLine();
	    lineCmd = true;
	    break;

	  case PREVIOUS_LINE:
	    handlePreviousLine();
	    lineCmd = true;
	    break;

	  case DUMP_LINES:
	    dumpLines();
	    break;

	  case DUMP_STYLES:
	    text.styles.dumpStyles();
	    break;

	  case SCROLL_LINE_UP:
	  case SCROLL_LINE_DOWN:
	    scrollLine(cmd.getCode() == SCROLL_LINE_UP ? -1 : 1);
	    break;

	  default:
	    if (!lineCmd)
		lineColumn = -1;
	}
	if (!lineCmd)
	    lineColumn = -1;
    }	

    void handleNextLine() {
	int caret = getCaret();

	if (lineColumn == -1) {
	    lineColumn = text.getColumnFor(caret);
	}
	int eol = text.getNextLine(caret);
	if (eol > 0) {
	    setCaret(text.getPosFor(lineColumn, eol));
	}
    }

    void handlePreviousLine() {
	int caret = getCaret();

	if (lineColumn == -1) {
	    lineColumn = text.getColumnFor(caret);
	}
	setCaret(text.getPosFor(lineColumn,
				text.getBeginningOfLine(caret) - 1));
    }

    /** This applies a style change to the current region. */
    public void applyStyleChange(StyleChange sc) {
	applyStyleChange(sc, positions[SELECTION_START],
			 positions[SELECTION_END]);
    }

    /** This applies a style change to the specified region. */
    public void applyStyleChange(StyleChange sc, int pos0, int pos1) {
	text.applyStyleChange(sc, pos0, pos1);
    }

    /**
     * Layout the lines, ensuring (if requested) that the caret is
     * visible when this redisplay is done.
     */
    final boolean performLayout() {
	boolean copy = layoutLines();

	if (makePosVisible >= 0 &&
	    !newLines.posVisible(makePosVisible, this)) {
	    scrollTo(makePosVisible, height / 2);
	    copy = layoutLines();
	}
	makePosVisible = -1;

	return copy;
    }

    private void initCaret() {
	if (!caretVisible) {
	    /* tick will toggle it on */
	    tick(0, null);
	}
	if (valid && isSet(F_EDITABLE) || (editor != null)) {
	    Timer.master.add(this, System.currentTimeMillis() + 500, null);
	}
    }

    public long tick(long tm, Object arg) {
	if (arg != null) {
	    synchronized (this) {
		/* autoscrolling */
		if (autoScrolling && autoScroll()) {
		    return tm + 50;
		}
		/* stop scrolling */
		autoScrolling = false;
	    }
	} else {
	    if (isSet(F_EDITABLE) || (editor != null)) {
		if (selectionLength() == 0) {
		    caretVisible = !caretVisible;
		    int pos = getCaret();
		    repaintRegion(pos, pos, false);
		}
		if (hasFocus()) {
		    return tm + 500;
		}
	    } else {
		caretVisible = true;
	    }
	}
	return -1;
    }

    public void paint(Graphics g) {
	if (!transparent) {
	    g.setColor(hilite);
	    g.fillRect(0, 0, width, height);
	}
	if (isSet(F_BOXED)) {
	    g.setColor(parent.background);
	    Bevel.drawFieldBorder(g, 0, 0, width, height);
	}
	if (visibleLines.lineCount == 0) {
	    /* never been painted before */
	    update(g);
	} else if (isSet(F_MODIFIED) ||
		   (makePosVisible >= 0 &&
		    !visibleLines.posVisible(makePosVisible, this))) {
	    /* we're not opaque - repaint whole thing every time */
	    synchronized (text) {
		synchronized (this) {
		    //System.out.println("Layout (paint)");
		    performLayout();

		    LinePool tmp = visibleLines;
		    visibleLines = newLines;
		    newLines = tmp;

		    paintLines(g, true);
		    clearFlag(F_MODIFIED);
		}
	    }
	} else {
	    //System.out.println("Just painting (paint)");
	    synchronized (text) {
		paintLines(g, true);
	    }
	}
    }

    /**
     * Incremental redisplay of the screen.  Just paints lines which
     * have changed.
     */
    public void update(Graphics g) {
	if (isSet(F_MODIFIED) ||
	    (makePosVisible >= 0 &&
	     !visibleLines.posVisible(makePosVisible, this))) {
	    //System.out.println("Update called");
	    synchronized (text) {
		synchronized (this) {
		    if (performLayout() && !transparent) {
			copyLines(g, 0, 0);
		    }
		    LinePool tmp = visibleLines;
		    visibleLines = newLines;
		    newLines = tmp;
		    paintLines(g, false);
		    clearFlag(F_MODIFIED);
		}
	    }
	} else {
	    //System.out.println("Just painting (update)");
	    synchronized (text) {
		paintLines(g, false);
	    }
	}
    }

/****************** scrolling *********************/

    int yvals[] = new int[10];
    int posvals[] = new int[10];

    /**
     * Try to get the line containing pos positioned at Y in the
     * window.  This works by backing up from destpos to line
     * bounderies, formatting them until "y" pixels worth of lines are
     * formatted.
     */
    protected void scrollTo(int destpos, int y) {
	if (!isSet(F_SCROLLING_OK))
	    return;

	Text text = this.text;
	Line line = new Line();
	StylePool styles = text.styles;
	int poslimit = text.length();
	int top = 0;

	while (destpos > 0) {
	    int start = text.scan('\n', destpos, -1);
	    int pos = start + 1;
		
	    styleIndex = styles.findStyleIndexFor(pos);
	    int lineCount = 0;
	    int height = 0;
	    while (pos <= destpos) {
		if (lineCount >= yvals.length) {
		    int tmp[] = new int[yvals.length + 10];
		    System.arraycopy(yvals, 0, tmp, 0, yvals.length);
		    yvals = tmp;
		    tmp = new int[posvals.length + 10];
		    System.arraycopy(posvals, 0, tmp, 0, posvals.length);
		    posvals = tmp;
		}
		Line l = visibleLines.findLineAt(pos);
		int last = pos;

		posvals[lineCount] = pos;
		if (l == null) {
		    l = line;
		    Style style = styles.getStyleAt(styleIndex);
		    float xlimit = width - style.rightMargin - insets.right;
		    l.pos = pos;
		    while (pos >= styles.getStylePosAt(styleIndex + 1))
			styleIndex += 1;
		    pos = measureLine(l, xlimit, poslimit);
		} else {
		    pos = l.pos1;
		}
		yvals[lineCount] = l.height;
		lineCount += 1;
		if (last == pos)
		    break;
	    }
	    while (--lineCount >= 0 && (y -= yvals[lineCount]) > 0)
		;
	    if (lineCount >= 0) {
		top = posvals[lineCount];
		break;
	    }
	    destpos = start;
	}
	setTop(top);
    }

    public Dimension preferredSize() {
	Text text = this.text;
	Line line = new Line();
	StylePool styles = text.styles;
	int pos = 0;
	int limit = text.length();
	int top = 0;
	int y = insets.top;

	validate();
	styleIndex = styles.findStyleIndexFor(0);

	while (pos < limit) {
	    Style style = styles.getStyleAt(styleIndex);
	    line.pos = pos;
	    float xlimit = width - style.rightMargin - insets.right;
	    pos = measureLine(line, xlimit, limit);
	    y += line.height;
	}
	return new Dimension(width, y + insets.bottom);
    }

    public void scrollAbsolute(int pos, int y) {
	scrollTo(pos, y);
    }

    public void scrollLine(int dir) {
	if (!isSet(F_SCROLLING_OK))
	    return;
	if (dir > 0) {
	    if (visibleLines.lineCount > 1)
		setTop(visibleLines.getLineAt(1).pos);
	} else {
	    Line first = visibleLines.getLineAt(0);
	    if (first.pos > 0)
		scrollTo(first.pos - 1, 0);
	}
    }

    public void scrollPage(int dir) {
	scrollPage(dir, false);
    }

    public void scrollPage(int dir, boolean moveCaret) {
	if (!isSet(F_SCROLLING_OK))
	    return;
	Line line;
	if (dir > 0) {
	    if (visibleLines.lineCount > 1) {
		line = visibleLines.getLineAt(visibleLines.lineCount - 1);
		if (line.pos1 < text.length()) {
		    setTop(line.pos);
		}
	    }
	} else {
	    if (visibleLines.lineCount > 0) {
		line = visibleLines.getLineAt(0);
		scrollTo(line.pos, height - line.height);
	    }
	}
	if (moveCaret) {
	    setCaret(positions[TOP]);
	}
    }

/************* Mouse handling. *****************/

    protected int pivot;
    protected int level;

    public boolean mouseDown(Event e, int x, int y) {
	if (!e.metaDown()) {
	    requestFocus();
	    level = e.clickCount;
	    if (e.shiftDown()) {
		/* if shift is down, then adjust existing selection */
		int pos = locateHit(x, y);
		int end = positions[SELECTION_END];
		int start = positions[SELECTION_START];
		pivot = (pos < (start + end) / 2) ? end : start;
		initCaret();
		return mouseDrag(e, x, y);
	    } else {
		pivot = locateHit(x, y);
		select(pivot, pivot, level);
		initCaret();
	    }
	}
	return false;
    }

    private boolean autoScrolling;
    private int sx, sy;

    public boolean mouseDrag(Event e, int x, int y) {
	int pos = locateHit(x, y);
	if (isSet(F_SELECTIONS_OK) || editor != null) {
	    select(pivot, pos, level);
	    if ((y < 0 && !visibleLines.posVisible(0, this)) ||
		(y > height - insets.bottom &&
		 !visibleLines.posVisible(text.length(), this))) {
		synchronized (this) {
		    if (!autoScrolling) {
			Timer.master.add(this, System.currentTimeMillis(), this);
			autoScrolling = true;
		    }
		}
		sx = x;
		sy = y;
	    } else {
		synchronized (this) {
		    if (autoScrolling) {
			autoScrolling = false;
			Timer.master.remove(this, this);
		    }
		}
	    }
	}
	return false;
    }

    public boolean mouseUp(Event e, int x, int y) {
	initCaret();
	synchronized (this) {
	    if (autoScrolling) {
		autoScrolling = false;
		Timer.master.remove(this, this);
	    }
	}
	return false;
    }

    /** Run the autoscrolling of selections. */
    public boolean autoScroll() {
	int y = sy;

	if (y < 0) {
	    scrollTo(positions[TOP], 16 + -y);
	} else if (y > height - insets.bottom) {
	    int diff = y - (height - insets.bottom);
	    scrollTo(locateHit(0, diff + 16), insets.top + 5);
	}
	boolean scrolled = isSet(F_MODIFIED);
	repaint();
	/* this causes the selection to be extended to the
	   newly exposed area */
	mouseDrag(null, sx, sy);
	return scrolled;
    }

    void dumpLines() {
	int i = 0;

	while (i < visibleLines.lineCount || i < newLines.lineCount) {
	    System.out.println("Line " + i);
	    if (i < visibleLines.lineCount) {
		Line line = visibleLines.getLineAt(i);
		if (line.pos >= 0)
		    System.out.println("V:\t" + line + "\n    \"" +
				       text.substring(line.pos, line.posV)
				       + "\"");
	    }

	    if (i < newLines.lineCount) {
		Line line = newLines.getLineAt(i);
		if (line.pos >= 0)
		    System.out.println("N:\t" + line + "\n    \"" +
				       text.substring(line.pos, line.posV)
				       + "\"");
	    }
	    i += 1;
	}
    }

/************************* KEYMAP STUFF **************************/

    /** The default keymap to use. */
    protected static Keymap defaultKeymap;

    /********** Commands this view can execute. ***********/
    public final static int BACKWARD_CHAR = 1;
    public final static int BACKWARD_WORD = 2;
    public final static int BEGINNING_OF_LINE = 3;
    public final static int END_OF_LINE = 4;
    public final static int FORWARD_CHAR = 5;
    public final static int FORWARD_WORD = 6;
    public final static int KILL_TO_EOL = 7;
    public final static int NEXT_PAGE = 9;
    public final static int PREVIOUS_PAGE = 11;
    public final static int SELF_INSERT = 12;
    public final static int DELETE_PREVIOUS_CHAR = 13;
    public final static int DELETE_NEXT_CHAR = 14;
    public final static int BEGINNING_OF_BUFFER = 15;
    public final static int END_OF_BUFFER = 16;
    public final static int DELETE_NEXT_WORD = 17;
    public final static int DELETE_PREVIOUS_WORD = 18;
    public final static int MAKE_ITALICS = 19;
    public final static int MAKE_BOLD = 20;
    public final static int MAKE_EMBOSS = 21;
    public final static int MAKE_LARGER = 22;
    public final static int MAKE_SMALLER = 23;
    public final static int MAKE_PLAIN = 24;
    public final static int COPY = 25;
    public final static int CUT = 26;
    public final static int PASTE = 27;
    public final static int NEWLINE_AND_BACKUP = 28;
    public final static int SET_MARK = 29;
    public final static int KILL_REGION = 30;
    public final static int NEXT_FOCUS = 31;
    public final static int END_OF_VIEW_LINE = 32;
    public final static int BEGINNING_OF_VIEW_LINE = 33;
    public final static int NEXT_LINE = 34;
    public final static int PREVIOUS_LINE = 35;
    public final static int FORWARD_WORD_START = 36;

    static final int DUMP_LINES = 1001;
    static final int SCROLL_LINE_DOWN = 1002;
    static final int SCROLL_LINE_UP = 1003;
    static final int TOGGLE_TRANSPARENT = 1004;
    static final int DUMP_STYLES = 1008;

    static Hashtable cmds = new Hashtable(64);

    public static Options alignOptions = new Options();
    static {
	alignOptions.add("left", LEFT);
	alignOptions.add("right", RIGHT);
	alignOptions.add("center", CENTER);
	alignOptions.add("fill", FILL);
    }

    public static Options wrapOptions = new Options();
    static {
	wrapOptions.add("char", Style.CHAR_WRAP);
	wrapOptions.add("word", Style.WORD_WRAP);
	wrapOptions.add("truncate", Style.TRUNCATE);
    }

    protected static KeyboardCommand lookupCmd(String name) {
	KeyboardCommand cmd = (KeyboardCommand) cmds.get(name);
	if (cmd == null) {
	    System.out.println("Warning: command not found " + name);
	}
	return cmd;
    }

    protected static KeyboardCommand getCommand(String name, int code) {
	KeyboardCommand cmd = (KeyboardCommand) cmds.get(name);

	if (cmd == null) {
	    cmd = new KeyboardCommand(name, code);
	    cmds.put(name, cmd);
	} else if (((KeyboardCommand) cmd).getCode() != code)
	    throw new RuntimeException(cmd + " cannot be reset to " + code);
	return cmd;
    }

    static {
	defaultKeymap = Keymap.getCombinationKeymap(null);
	KeyboardCommand insert;
	insert = getCommand("self-insert", SELF_INSERT);
	for (int c = ' '; c < 127; c++)
	    defaultKeymap.setKey(c, insert);
	defaultKeymap.setKey('\t', insert);
	defaultKeymap.bindKeys("\\Control-J", insert);
	defaultKeymap.setKey('\n', insert);
	defaultKeymap.bindKeys("\\C-@", getCommand("set-mark",
						   SET_MARK));
	defaultKeymap.bindKeys("\\C-a", getCommand("beginning-of-line",
						   BEGINNING_OF_LINE));
	defaultKeymap.bindKeys("\\C-c", getCommand("copy-to-clipboard",
						   COPY));
	defaultKeymap.bindKeys("\\C-d", getCommand("delete-next-character",
						   DELETE_NEXT_CHAR));
	defaultKeymap.setKey('\177', lookupCmd("delete-next-character"));
	defaultKeymap.bindKeys("\\C-e", getCommand("end-of-line",
						   END_OF_LINE));
	defaultKeymap.bindKeys("\\C-f", getCommand("forward-character",
						   FORWARD_CHAR));
	defaultKeymap.bindKeys("\b", getCommand("delete-previous-character",
						DELETE_PREVIOUS_CHAR));
	defaultKeymap.bindKeys("\\C-o", getCommand("newline-and-backup",
						   NEWLINE_AND_BACKUP));
	defaultKeymap.bindKeys("\\C-x", getCommand("cut-to-clipboard",
						   CUT));
	defaultKeymap.bindKeys("\\C-w", getCommand("kill-region",
						   KILL_REGION));
	defaultKeymap.bindKeys("\\C-k", getCommand("kill-to-end-of-line",
						   KILL_TO_EOL));
	defaultKeymap.bindKeys("\\C-v", getCommand("paste-clipboard",
						   PASTE));
	defaultKeymap.bindKeys("\\M-b", getCommand("backward-word",
						   BACKWARD_WORD));
	defaultKeymap.bindKeys("\\M-d", getCommand("delete-next-word",
						   DELETE_NEXT_WORD));
	defaultKeymap.bindKeys("\\M-f", getCommand("forward-word",
						   FORWARD_WORD));
	defaultKeymap.bindKeys("\\M-h", getCommand("delete-previous-word",
						   DELETE_PREVIOUS_WORD));
	defaultKeymap.bindKeys("\\M-\177", getCommand("delete-previous-word",
						      DELETE_PREVIOUS_WORD));
	defaultKeymap.bindKeys("\\M-<", getCommand("beginning-of-buffer",
						   BEGINNING_OF_BUFFER));
	defaultKeymap.bindKeys("\\M->", getCommand("end-of-buffer",
						   END_OF_BUFFER));
	defaultKeymap.bindKeys("\\Control-I", getCommand("make-italics",
							 MAKE_ITALICS));
	defaultKeymap.bindKeys("\\Control-B", getCommand("make-bold",
							 MAKE_BOLD));
	defaultKeymap.bindKeys("\\Control-S\\Control-E",
			       getCommand("make-emboss", MAKE_EMBOSS));
	defaultKeymap.bindKeys("\\Control-S\\Control-L",
			       getCommand("make-larger", MAKE_LARGER));
	defaultKeymap.bindKeys("\\Control-S\\Control-S",
			       getCommand("make-smaller", MAKE_SMALLER));
	defaultKeymap.bindKeys("\\Control-S\\Control-P",
			       getCommand("make-plain", MAKE_PLAIN));
	defaultKeymap.bindKeys("\\Left", getCommand("backward-character",
						    BACKWARD_CHAR));
	defaultKeymap.bindKeys("\\Right", lookupCmd("forward-character"));
	defaultKeymap.bindKeys("\\Home", getCommand("beginning-of-view-line",
						    BEGINNING_OF_VIEW_LINE));
	defaultKeymap.bindKeys("\\End", getCommand("end-of-view-line",
						    END_OF_VIEW_LINE));
	defaultKeymap.bindKeys("\\Control-\\End", lookupCmd("end-of-buffer"));
	defaultKeymap.bindKeys("\\Control-\\Home",
			       lookupCmd("beginning-of-buffer"));
	defaultKeymap.bindKeys("\\Control-\\Right", getCommand("forward-word-start", FORWARD_WORD_START));
	defaultKeymap.bindKeys("\\Control-\\Left", lookupCmd("backward-word"));
	defaultKeymap.bindKeys("\\PGUP", getCommand("previous-page", PREVIOUS_PAGE));
	defaultKeymap.bindKeys("\\Pgdn", getCommand("next-page", NEXT_PAGE));
	defaultKeymap.bindKeys("\\Down", getCommand("next-line", NEXT_LINE));
	defaultKeymap.bindKeys("\\Up", getCommand("previous-line",
						  PREVIOUS_LINE));
	defaultKeymap.bindKeys("\\M-\b", lookupCmd("delete-previous-word"));
	defaultKeymap.bindKeys("\\C-n", lookupCmd("next-line"));
	defaultKeymap.bindKeys("\\C-p", lookupCmd("previous-line"));
	defaultKeymap.bindKeys("\\Meta-^", getCommand("dump-line-info",
						      DUMP_LINES));
	defaultKeymap.setKey(Event.F1, getCommand("dump-styles",
						  DUMP_STYLES));
    }
}
